# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Main module to open the QT UI."""

import os
import sys

from PyQt4 import QtGui, QtCore

# Module used to include the resources into this file
from ubuntu_kylin_sso.qt.ui import resources_rc
from ubuntu_kylin_sso.qt.ubuntu_sso_wizard import UbuntuSSOClientGUI
from ubuntu_kylin_sso.utils import compat, PLATFORM_QSS


# Poke resources_rc to avoid pyflakes complaining about unused import
assert(resources_rc)

if sys.platform in ('win32', 'darwin'):
    from ubuntu_kylin_sso.qt.main import windows
    source = windows
else:
    from ubuntu_kylin_sso.qt.main import linux
    source = linux


def main(**kwargs):
    """Start the QT mainloop and open the main window."""
    if os.environ.get('TESTABILITY', False) and \
            '-testability' not in sys.argv:
        sys.argv.append('-testability')
    app = QtGui.QApplication(sys.argv)

    source.main(app)

    data = []
    for qss_name in (PLATFORM_QSS, ":/stylesheet.qss"):
        qss = QtCore.QResource(qss_name)
        data.append(compat.text_type(qss.data()))
    app.setStyleSheet('\n'.join(data))

    # Fix the string that contains unicode chars.
    for key in kwargs:
        value = kwargs[key]
        if isinstance(value, compat.binary_type):
            kwargs[key] = value.decode('utf-8')

    close_callback = lambda: source.main_quit(app)
    ui = UbuntuSSOClientGUI(close_callback=close_callback, **kwargs)
    style = QtGui.QStyle.alignedRect(
                    QtCore.Qt.LeftToRight, QtCore.Qt.AlignCenter,
                    ui.size(), app.desktop().availableGeometry())
    ui.setGeometry(style)

    app = QtGui.QApplication.instance()
    # app.setWindowIcon(QtGui.QIcon.fromTheme("ubuntukylin"))
    app.setWindowIcon(QtGui.QIcon(QtGui.QPixmap('/usr/share/app-install/icons/youker-id.png')))

    ui.show()
    if sys.platform == 'darwin':
        ui.raise_()

    source.main_start(app)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""QArrow tests."""

from PyQt4.QtGui import QStyle
from twisted.internet import defer
from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso.qt.arrow import QArrow

# pylint: disable=W0212,C0103


class FakeStyle(object):
    """Fake a Qt style."""

    def __init__(self):
        """Create instance."""
        self.called = []

    def __call__(self, *args, **kwargs):
        """Instance callable."""
        return self

    def drawPrimitive(self, primitive, opt, painter, widget):
        """Fake a draw primitive."""
        self.called.append(('drawPrimitive', primitive, opt, painter, widget))


class QArrowTestCase(TestCase):
    """Test the QArrow class."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the tests."""
        yield super(QArrowTestCase, self).setUp()
        self.valid_dirs = (QArrow.UP, QArrow.DOWN,
                           QArrow.LEFT, QArrow.RIGHT)
        self.style = FakeStyle()

    def test_correct_init(self):
        """Test the init using the correct direction."""
        for direction in self.valid_dirs:
            arrow = QArrow(direction)
            self.assertEqual(direction, arrow.direction)

    def test_wrong_init(self):
        """Test the init using the wrong direction."""
        self.assertRaises(ValueError, QArrow, 'test')

    def test_set_direction_correct(self):
        """Test the set direction with the correct value."""
        arrow = QArrow(QArrow.RIGHT)
        for direction in self.valid_dirs:
            arrow.direction = direction
            self.assertEqual(direction, arrow.direction)

    def test_set_direction_wrong(self):
        """Test set direction wrong."""
        arrow = QArrow(QArrow.RIGHT)
        self.assertRaises(ValueError, setattr, arrow, 'direction', 'wrong')

    def test_paint_event(self):
        """Test the paint event."""
        directions_map = {QArrow.UP: QStyle.PE_IndicatorArrowUp,
            QArrow.DOWN: QStyle.PE_IndicatorArrowDown,
            QArrow.LEFT: QStyle.PE_IndicatorArrowLeft,
            QArrow.RIGHT: QStyle.PE_IndicatorArrowRight}

        arrow = QArrow(QArrow.RIGHT)
        self.patch(arrow, 'style', self.style)

        for key, value in directions_map.items():
            self.style.called = []
            arrow.direction = key
            arrow.paintEvent(None)
            self.assertIn(value, self.style.called[0])



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""tcpactivation: start a process if nothing listening in a given port."""

import subprocess

from twisted.internet import defer, error, protocol
from twisted.internet.endpoints import clientFromString

LOCALHOST = "127.0.0.1"
DELAY_BETWEEN_CHECKS = 0.1
NUMBER_OF_CHECKS = 600

# twisted uses a different coding convention
# pylint: disable=C0103,W0232


def async_sleep(delay):
    """Fire the returned deferred after some specified delay."""
    from twisted.internet import reactor
    d = defer.Deferred()
    # pylint: disable=E1101
    reactor.callLater(delay, d.callback, None)
    return d


class AlreadyStartedError(Exception):
    """The instance was already started."""


class ActivationTimeoutError(Exception):
    """Timeout while trying to start the instance."""


class NullProtocol(protocol.Protocol):
    """A protocol that drops the connection."""

    def connectionMade(self):
        """Just drop the connection."""
        self.transport.loseConnection()


class PortDetectFactory(protocol.ClientFactory):
    """Will detect if something is listening in a given port."""

    protocol = NullProtocol

    def __init__(self):
        """Initialize this instance."""
        self.d = defer.Deferred()

    def is_listening(self):
        """A deferred that will become True if something is listening."""
        return self.d

    def buildProtocol(self, addr):
        """Connected."""
        p = protocol.ClientFactory.buildProtocol(self, addr)
        if not self.d.called:
            self.d.callback(True)
        return p

    def clientConnectionLost(self, connector, reason):
        """The connection was lost."""
        protocol.ClientFactory.clientConnectionLost(self, connector, reason)
        if not self.d.called:
            self.d.callback(False)

    def clientConnectionFailed(self, connector, reason):
        """The connection failed."""
        protocol.ClientFactory.clientConnectionFailed(self, connector, reason)
        if not self.d.called:
            self.d.callback(False)


class ActivationConfig(object):
    """The configuration for tcp activation."""

    def __init__(self, service_name, command_line, description):
        """Initialize this instance."""
        self.service_name = service_name
        self.command_line = command_line
        self.description = description


class ActivationDetector(object):
    """Base class to detect if the service is running."""

    def __init__(self, config):
        """Initialize this instance."""
        self.config = config

    @defer.inlineCallbacks
    def is_already_running(self):
        """Check if the instance is already running."""
        from twisted.internet import reactor
        factory = PortDetectFactory()
        client = clientFromString(reactor, self.config.description.client)
        try:
            yield client.connect(factory)
        except error.ConnectError:
            defer.returnValue(False)
        result = yield factory.is_listening()
        defer.returnValue(result)


class ActivationClient(ActivationDetector):
    """A client for tcp activation."""

    # a classwide lock, so the server is started only once
    lock = defer.DeferredLock()

    @defer.inlineCallbacks
    def _wait_server_active(self):
        """Wait till the server is active."""
        for _ in range(NUMBER_OF_CHECKS):
            is_running = yield self.is_already_running()
            if is_running:
                defer.returnValue(None)
            yield async_sleep(DELAY_BETWEEN_CHECKS)
        raise ActivationTimeoutError()

    def _spawn_server(self):
        """Start running the server process."""
        # Without using close_fds=True, strange things happen
        # with logging on windows. More information at
        # http://bugs.python.org/issue4749
        subprocess.Popen(self.config.command_line, close_fds=True)

    @defer.inlineCallbacks
    def _do_get_active_description(self):
        """Get the details for the running instance, starting it if needed."""
        is_running = yield self.is_already_running()
        if not is_running:
            self._spawn_server()
            yield self._wait_server_active()
        defer.returnValue(self.config.description.client)

    @defer.inlineCallbacks
    def get_active_client_description(self):
        """Serialize the requests to _do_get_active_description."""
        yield self.lock.acquire()
        try:
            result = yield self._do_get_active_description()
            defer.returnValue(result)
        finally:
            self.lock.release()


class ActivationInstance(ActivationDetector):
    """A tcp activation server instance."""

    @defer.inlineCallbacks
    def get_server_description(self):
        """Get the port to run this service or fail if already started."""
        is_running = yield self.is_already_running()
        if is_running:
            raise AlreadyStartedError()
        defer.returnValue(self.config.description.server)



# -*- coding: utf-8 -*-
#
# Copyright 2009-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the Ubuntu SSO library."""

from __future__ import unicode_literals

import os

from collections import defaultdict
from functools import wraps

from twisted.internet import defer
from twisted.trial import unittest

from ubuntu_kylin_sso.keyring import get_token_name

APP_NAME = 'Ubuntu One'
CAPTCHA_ID = 'test ñiña'
CAPTCHA_PATH = os.path.abspath(os.path.join(os.curdir, 'ubuntu_kylin_sso', 'tests',
                                            'files', 'captcha.png'))
CAPTCHA_SOLUTION = 'william Byrd ñandú'
EMAIL = 'test@example.com'
EMAIL_TOKEN = 'B2P☺ gtf'
GUI_EXE = 'ubuntu-sso-login-qt'
HELP_TEXT = ('☛ Lorem ipsum dolor sit amet, consectetur adipiscing elit. '
             'Nam sed lorem nibh. Suspendisse gravida nulla non nunc suscipit '
             'pulvinar tempus ut augue. Morbi consequat, ligula a elementum '
             'pretium, dolor nulla tempus metus, sed viverra nisi risus non '
             'velit.')
NAME = 'Juanito ☀ Pérez'
PASSWORD = 'h3lloWorld☑ '
PING_URL = 'http://localhost/ping-me/'
POLICY_URL = 'http://localhost/policy/'
RESET_PASSWORD_TOKEN = '8G5Wtq'
TOKEN = {'consumer_key': 'xQ7xDAz',
         'consumer_secret': 'KzCJWCTNbbntwfyCKKjomJDzlgqxLy',
         'token_name': 'test',
         'token': 'GkInOfSMGwTXAUoVQwLUoPxElEEUdhsLVNTPhxHJDUIeHCPNEo',
         'token_secret': 'qFYImEtlczPbsCnYyuwLoPDlPEnvNcIktZphPQklAWrvyfFMV'}
TOKEN_NAME = get_token_name(APP_NAME)
TC_URL = 'http://localhost/'
WINDOW_ID = 5


class TestCase(unittest.TestCase):
    """Customized test case that keeps tracks of method calls."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(TestCase, self).setUp()
        self._called = False

    def _set_called(self, *args, **kwargs):
        """Keep track of a method call."""
        self._called = (args, kwargs)


class Recorder(object):
    """A class that records every call clients made to it."""

    no_wrap = ['_called']
    _next_result = None

    def __init__(self, *args, **kwargs):
        self._called = defaultdict(list)

    def __getattribute__(self, attr_name):
        """Override so we can record calls to members."""
        try:
            result = super(Recorder, self).__getattribute__(attr_name)
        except AttributeError:
            result = lambda *a, **kw: self._next_result
            super(Recorder, self).__setattr__(attr_name, result)

        if attr_name in super(Recorder, self).__getattribute__('no_wrap'):
            return result

        called = super(Recorder, self).__getattribute__('_called')

        def wrap_me(f):
            """Wrap 'f'."""
            @wraps(f)
            def inner(*a, **kw):
                """Keep track of calls to 'f', execute it and return result."""
                called[attr_name].append((a, kw))
                return f(*a, **kw)

            return inner

        if callable(result):
            return wrap_me(result)
        else:
            return result



# -*- coding: utf-8 -*-
#
# Copyright 2012-2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Qt specific tests for the main module."""

# pylint: disable=F0401
from PyQt4 import QtCore
from twisted.internet import defer

from ubuntu_kylin_sso.main import qt
from ubuntu_kylin_sso.tests import TestCase


class MiscTestCase(TestCase):
    """Tests for module level misc functions."""

    def clean_timers(self):
        """Remove all timers."""
        for timer in qt.TIMERS:
            timer.stop()
        qt.TIMERS = set()

    @defer.inlineCallbacks
    def setUp(self):
        yield super(MiscTestCase, self).setUp()
        self.addCleanup(self.clean_timers)

    def test_timeout_func_ignores_priority(self):
        """Should not pass priority as argument."""
        qt.timeout_func(0, self._set_called, priority="crazy")
        QtCore.QCoreApplication.instance().processEvents()
        self.assertEqual(self._called, ((), {}))

    def test_timeout_func_keeps_reference(self):
        """Should keep a reference to the timer created."""
        self.assertEqual(len(qt.TIMERS), 0)
        qt.timeout_func(0, self._set_called, priority="crazy")
        self.assertEqual(len(qt.TIMERS), 1)

    def test_timeout_func_removes_reference(self):
        """Remove reference to timer when callback returns False."""
        self.assertEqual(len(qt.TIMERS), 0)
        qt.timeout_func(0, self._set_called, priority="crazy")
        self.assertEqual(len(qt.TIMERS), 1)
        QtCore.QCoreApplication.instance().processEvents()
        self.assertEqual(len(qt.TIMERS), 0)

    def test_timeout_func_timer_activation(self):
        """The timer should be activated with the right period."""
        self.assertEqual(len(qt.TIMERS), 0)
        qt.timeout_func(1337, self._set_called, priority="crazy")
        self.assertEqual(len(qt.TIMERS), 1)
        self.assertEqual(list(qt.TIMERS)[0].interval(), 1337)

    def test_timeout_func_argument_passing(self):
        """The right arguments should be passed to the callback."""
        qt.timeout_func(0, self._set_called, 1, name="a")
        QtCore.QCoreApplication.instance().processEvents()
        self.assertEqual(self._called, ((1,), {"name": "a"}))

    def test_start_setup_calls_turkish_fix(self):
        """The start_setup function calls the turkish fix function."""
        self.patch(qt.QtCore, "QCoreApplication",
                   lambda *args, **kwargs: object())
        self.patch(qt.dbus.mainloop.qt, "DBusQtMainLoop",
                   lambda **kwargs: None)
        called = []
        self.patch(qt, "fix_turkish_locale", lambda: called.append(1))

        qt.start_setup()
        self.assertEqual(called, [1])



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Utility to spawn another program from a plain Qt mainloop."""

from PyQt4 import QtCore

from ubuntu_kylin_sso.logger import setup_logging


logger = setup_logging("ubuntu_kylin_sso.utils.runner.qt")

# pylint: disable=C0103
# the set below is a workaround to hold to object references to avoid
# garbage collection before the processes finish executing
_processes = set()


def spawn_program(args, reply_handler, error_handler):
    """Spawn the program specified by 'args' using the Qt mainloop.

    When the program finishes, 'reply_handler' will be called with a single
    argument that will be the porgram status code.

    If there is an error, error_handler will be called with an instance of
    SpawnError.

    """

    process = QtCore.QProcess()
    _processes.add(process)

    def print_pid():
        """Add a debug log message."""
        pid = process.pid()
        logger.debug('Spawning the program %r with the qt mainloop '
                     '(returned pid is %r).', args, pid)

    def child_watch(status):
        """Handle child termination."""
        reply_handler(status)
        _processes.remove(process)

    def handle_error(process_error):
        """Handle error when spawning the process."""
        failed_to_start = (process_error == process.FailedToStart)
        msg = 'ProcessError is %r' % process_error
        error_handler(msg=msg, failed_to_start=failed_to_start)
        _processes.remove(process)

    process.started.connect(print_pid)
    process.finished.connect(child_watch)
    process.error.connect(handle_error)

    args = list(args)
    program = args[0]
    argv = args[1:]
    process.start(program, argv)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utility classes for testing main on windows."""

from twisted.internet import defer

from ubuntu_kylin_sso import main
from ubuntu_kylin_sso.utils.tests.test_ipc import BaseIPCTestCase


class BaseTestCase(BaseIPCTestCase):
    """The base test case with platform specific support."""

    client_class = main.UbuntuSSOClient
    service_class = lambda *a: main.UbuntuSSOProxy(None)

    @defer.inlineCallbacks
    def setUp(self):

        self.sso_service = None

        def service_factory(*args):
            """Returns the service class."""
            self.sso_service = main.UbuntuSSOService()
            return main.UbuntuSSOProxy(self.sso_service)

        self.service_class = service_factory

        yield super(BaseTestCase, self).setUp()

        # set the proxy to be the service started by the parent test class
        self.sso_service.proxy = self.service

        # patch the start of the proxy since the parent setup started it
        self.patch(self.sso_service.proxy, 'start',
                                            lambda: defer.succeed(None))

        # start the service which will perform all the required operations
        # except the proxy start
        yield self.sso_service.start()

        self.sso_client = self.client



# -*- coding: utf-8 -*-
# Author: Diego Sarmentero <diego.sarmentero@canonical.com>
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test the common functions."""

from __future__ import unicode_literals

from PyQt4 import QtGui
from twisted.internet import defer
from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso.qt import (
    build_general_error_message,
    maybe_elide_text,
    GENERIC_BACKEND_ERROR,
)
from ubuntu_kylin_sso.qt.common import (
    check_as_invalid,
    check_as_valid,
    password_assistance,
    password_check_match,
    BAD,
    GOOD,
    NORMAL,
    PASSWORD_DIGIT,
    PASSWORD_LENGTH,
    PASSWORD_MATCH,
    PASSWORD_UPPER,
)
from ubuntu_kylin_sso.qt.tests import build_string_for_pixels


class PasswordTestCase(TestCase):
    """Test passwords conditions."""

    @defer.inlineCallbacks
    def setUp(self):
        """Setup tests."""
        yield super(PasswordTestCase, self).setUp()
        self.line_edit = QtGui.QLineEdit()
        self.line_edit_confirm = QtGui.QLineEdit()
        self.label_assistance = QtGui.QLabel()

    def test_short_password(self):
        """Status with short password.

        * Password assistance contains the right message.
        """
        self.line_edit.setText("foobar")
        password_assistance(self.line_edit, self.label_assistance)
        self.assertIn(
            BAD % PASSWORD_LENGTH,
            compat.text_type(self.label_assistance.text()),
        )

    def test_long_password(self):
        """Status with long password.

        * Password assistance contains the right message.
        """
        self.line_edit.setText("foobarbaz")
        password_assistance(self.line_edit, self.label_assistance)
        self.assertIn(
            GOOD % PASSWORD_LENGTH,
            compat.text_type(self.label_assistance.text()),
        )

    def test_no_number_password(self):
        """Status with password without a number.

        * Password assistance contains the right message.
        """
        self.line_edit.setText("foobarbaz")
        password_assistance(self.line_edit, self.label_assistance)
        self.assertIn(
            BAD % PASSWORD_DIGIT,
            compat.text_type(self.label_assistance.text()),
        )

    def test_number_password(self):
        """Status with password with a number.

        * Password assistance contains the right message.
        """
        self.line_edit.setText("foobarba7")
        password_assistance(self.line_edit, self.label_assistance)
        self.assertIn(
            GOOD % PASSWORD_DIGIT,
            compat.text_type(self.label_assistance.text()),
        )

    def test_no_uppercase_password(self):
        """Status with password without uppercase letters.

        * Password assistance contains the right message.
        """
        self.line_edit.setText("foobarbaz")
        password_assistance(self.line_edit, self.label_assistance)
        self.assertIn(
            BAD % PASSWORD_UPPER,
            compat.text_type(self.label_assistance.text()),
        )

    def test_upper_password(self):
        """Status with password with uppercase letters.

        * Password assistance contains the right message.
        """
        self.line_edit.setText("Foobarba7")
        password_assistance(self.line_edit, self.label_assistance)
        self.assertIn(
            GOOD % PASSWORD_UPPER,
            compat.text_type(self.label_assistance.text()),
        )

    def test_matching_passwords(self):
        """Status when the passwords match.

        * Password assistance doesn't contain the message.
        """
        self.line_edit.setText("Foobarba7")
        self.line_edit_confirm.setText("Foobarba7")
        password_check_match(self.line_edit,
            self.line_edit_confirm, self.label_assistance)
        self.assertNotIn(
            BAD % PASSWORD_MATCH,
            compat.text_type(self.label_assistance.text()))

    def test_not_matching_passwords(self):
        """Status when the passwords not match.

        * Password assistance contains the right message.
        """
        self.line_edit.setText("Foobarba7")
        self.line_edit_confirm.setText("BazBarFo0")
        password_check_match(self.line_edit,
            self.line_edit_confirm, self.label_assistance)
        self.assertIn(
            BAD % PASSWORD_MATCH,
            compat.text_type(self.label_assistance.text()),
        )

    def test_password_assistance_in_focus_length_correct(self):
        """Check the QLineEdit for password when the length is correct."""
        self.line_edit.setText("aaaaaaaaa")
        self.line_edit_confirm.setText("")
        password_assistance(self.line_edit, self.label_assistance, NORMAL)
        self.assertIn(
            GOOD % PASSWORD_LENGTH,
            compat.text_type(self.label_assistance.text()),
        )
        self.assertIn(
            NORMAL % PASSWORD_UPPER,
            compat.text_type(self.label_assistance.text()),
        )
        self.assertIn(
            NORMAL % PASSWORD_DIGIT,
            compat.text_type(self.label_assistance.text()),
        )

    def test_password_assistance_in_focus_with_upper(self):
        """Check the QLineEdit for password when it has an upper case char."""
        self.line_edit.setText("AAa")
        self.line_edit_confirm.setText("")
        password_assistance(self.line_edit, self.label_assistance, NORMAL)
        self.assertIn(
            NORMAL % PASSWORD_LENGTH,
            compat.text_type(self.label_assistance.text()),
        )
        self.assertIn(
            GOOD % PASSWORD_UPPER,
            compat.text_type(self.label_assistance.text()),
        )
        self.assertIn(
            NORMAL % PASSWORD_DIGIT,
            compat.text_type(self.label_assistance.text()),
        )

    def test_password_assistance_in_focus_with_number(self):
        """Check the QLineEdit for password when it contains numbers."""
        self.line_edit.setText("a123")
        self.line_edit_confirm.setText("")
        password_assistance(self.line_edit, self.label_assistance, NORMAL)
        self.assertIn(
            NORMAL % PASSWORD_LENGTH,
            compat.text_type(self.label_assistance.text()),
        )
        self.assertIn(
            NORMAL % PASSWORD_UPPER,
            compat.text_type(self.label_assistance.text()),
        )
        self.assertIn(
            GOOD % PASSWORD_DIGIT,
            compat.text_type(self.label_assistance.text()),
        )

    def test_password_assistance_in_focus_all_ok(self):
        """Check the QLineEdit for password when all the condition are ok."""
        self.line_edit.setText("T3st3rqw")
        self.line_edit_confirm.setText("T3st3rqw")
        password_assistance(self.line_edit, self.label_assistance, NORMAL)
        self.assertIn(
            GOOD % PASSWORD_LENGTH,
            compat.text_type(self.label_assistance.text()),
        )
        self.assertIn(
            GOOD % PASSWORD_UPPER,
            compat.text_type(self.label_assistance.text()),
        )
        self.assertIn(
            GOOD % PASSWORD_DIGIT,
            compat.text_type(self.label_assistance.text()),
        )
        self.assertNotIn(
            NORMAL % PASSWORD_MATCH,
            compat.text_type(self.label_assistance.text()),
        )


class LineEditStyleTestCase(TestCase):
    """Test QLineEdit styles for errors."""

    def test_check_valid(self):
        """Check the propery value of a QLineEdit marked as valid."""
        line_edit = QtGui.QLineEdit()
        check_as_valid(line_edit)
        self.assertFalse(line_edit.property("formError").toBool())

    def test_check_invalid(self):
        """Check the propery value of a QLineEdit marked as invalid."""
        line_edit = QtGui.QLineEdit()
        check_as_invalid(line_edit)
        self.assertTrue(line_edit.property("formError").toBool())


class ElidedTextTestCase(TestCase):
    """The test case for the maybe_elide_text function."""

    max_width = 100

    @defer.inlineCallbacks
    def setUp(self):
        """Setup tests."""
        yield super(ElidedTextTestCase, self).setUp()
        self.ui = QtGui.QLabel()

    def test_text_not_elided_if_too_short(self):
        """If text is shorter than max_width, do not elide."""
        text = build_string_for_pixels(self.ui, self.max_width - 10)

        maybe_elide_text(self.ui, text, self.max_width)

        self.assertEqual(self.ui.toolTip(), '')
        self.assertEqual(self.ui.text(), text)
        self.assertNotIn('\u2026', self.ui.text())

    def test_text_not_elided_if_equals_max_width(self):
        """If text is equal than max_width, do not elide."""
        text = build_string_for_pixels(self.ui, self.max_width)

        maybe_elide_text(self.ui, text, self.max_width)

        self.assertEqual(self.ui.toolTip(), '')
        self.assertEqual(self.ui.text(), text)
        self.assertNotIn('\u2026', self.ui.text())

    def test_text_elided_if_bigger_than_max_width(self):
        """If text is equal than max_width, do not elide."""
        text = build_string_for_pixels(self.ui, self.max_width * 2)

        maybe_elide_text(self.ui, text, self.max_width)

        self.assertEqual(self.ui.toolTip(), text)
        expected = compat.text_type(self.ui.text())
        self.assertTrue(expected.endswith('\u2026'))
        self.assertTrue(text.startswith(expected[:-1]))


class BuildGeneralErrorMessageTestCase(TestCase):
    """Test passwords conditions."""

    def test_with_message(self):
        """Test build_general_error_message with 'message' key."""
        error = "error message"
        err_dict = {'message': error}

        result = build_general_error_message(err_dict)

        self.assertEqual(result, error)

    def test_with_all(self):
        """Test build_general_error_message with 'all' key."""
        error = "error message"
        err_dict = {'__all__': error}

        result = build_general_error_message(err_dict)

        self.assertEqual(result, error)

    def test_with_message_and_all(self):
        """Test build_general_error_message with 'all' and 'message' key."""
        error = "error message"
        error2 = "error message2"
        err_dict = {'__all__': error, 'message': error2}

        result = build_general_error_message(err_dict)

        expected = '\n'.join((error, error2))
        self.assertEqual(result, expected)

    def test_with_all_and_error_message(self):
        """Test for 'all' and 'error_message' key."""
        error = "error message"
        error2 = "error message2"
        err_dict = {'__all__': error, 'error_message': error2}
        result = build_general_error_message(err_dict)
        expected = '\n'.join((error, error2))
        self.assertEqual(result, expected)

    def test_with_random_keys(self):
        """Test build_general_error_message with random keys."""
        error = "error message"
        error2 = "error message2"
        err_dict = {'my_bad': error, 'odd_error': error2}

        result = build_general_error_message(err_dict)

        expected = '\n'.join(
            [('%s: %s' % (k, v)) for k, v in err_dict.items()])
        self.assertEqual(result, expected)

    def test_with_random_keys_with_errtype(self):
        """Test build_general_error_message with random keys and errtype."""
        error = "error message"
        error2 = "error message2"
        err_dict = {'my_bad': error, 'odd_error': error2, 'errtype': 'Danger'}

        result = build_general_error_message(err_dict)

        expected = '\n'.join(
            [('%s: %s' % (k, v))
            for k, v in {'my_bad': error, 'odd_error': error2}.items()])
        self.assertEqual(result, expected)

    def test_with_not_dict(self):
        """Test build_general_error_message with argument not dict."""
        error = "error message"
        err_dict = Exception(error)

        result = build_general_error_message(err_dict)

        expected = GENERIC_BACKEND_ERROR
        self.assertEqual(result, expected)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the main SSO client code."""

from twisted.internet import defer
from ubuntuone.devtools.testcases import skipIfOS

from ubuntu_kylin_sso import main
from ubuntu_kylin_sso.tests import (
    APP_NAME,
    CAPTCHA_ID,
    CAPTCHA_SOLUTION,
    EMAIL,
    EMAIL_TOKEN,
    NAME,
    PASSWORD,
    TOKEN,
)
from ubuntu_kylin_sso.main.tests import BaseTestCase, FakedCredentials

FILENAME = 'sample filename'


class FakedKeyring(object):
    """A faked Keyring object."""

    _keys = {}

    def get_credentials(self, app_name):
        """Return the credentials for app_name."""
        return defer.succeed(self._keys.get(app_name, {}))

    def delete_credentials(self, app_name):
        """Delete the credentials for app_name."""
        self._keys.pop(app_name, None)
        return defer.succeed(None)

    def set_credentials(self, app_name, token):
        """Store the credentials for app_name."""
        self._keys[app_name] = token
        return defer.succeed(None)


class AbstractTestCase(BaseTestCase):
    """The base test case with platform specific support."""

    timeout = 2
    method = None
    backend_method = method
    params = ()
    success_signal = None
    backend_result = None
    success_result = None
    error_signal = None

    @defer.inlineCallbacks
    def setUp(self):
        # avoid putting stuff in the mainloops this MUST be done
        # before the parent setUp is called because it will register
        # the callbacks
        self.patch(main.source, 'timeout_func', lambda *a: None)
        self.patch(main.source, 'shutdown_func', lambda *a: None)

        yield super(AbstractTestCase, self).setUp()
        self.keyring = FakedKeyring()
        self.patch(main, 'Keyring', lambda: self.keyring)

        self.test_client = self.patchable_backend = None

        if self.backend_method is None:
            self.backend_method = self.method

    def _backend_succeed(self, *args, **kwargs):
        """Make self.patchable_backend return self.backend_result."""
        return self.backend_result

    def _backend_fail(self, *args, **kwargs):
        """Make self.patchable_backend raise an exception."""
        raise ValueError((args, kwargs))

    @defer.inlineCallbacks
    def assert_method_correct(self, success_signal, error_signal,
                              patched_backend_method, expected_result=None):
        """Calling 'self.method' works ok.

        Check that 'success_signal' is emitted, and make the test fail if
        error_signal is received.

        The self.patchable_backend will be patched with
        'patched_backend_method'.

        """
        if self.method is None:
            defer.returnValue(None)

        d = defer.Deferred()

        cb = lambda *a: d.callback(a)
        match = self.test_client.connect_to_signal(success_signal, cb)
        self.addCleanup(self.test_client.disconnect_from_signal,
                        success_signal, match)

        eb = lambda *a: d.errback(AssertionError(a))
        match = self.test_client.connect_to_signal(error_signal, eb)
        self.addCleanup(self.test_client.disconnect_from_signal,
                        error_signal, match)

        self.patch(self.patchable_backend, self.backend_method,
                   patched_backend_method)

        yield self.test_client.call_method(self.method, *self.params)

        result = yield d
        self.assertEqual(expected_result, result)

    def test_success(self):
        """Test that the 'method' works ok."""
        success_signal = self.success_signal
        error_signal = self.error_signal
        patched_backend_method = self._backend_succeed
        expected_result = self.success_result

        return self.assert_method_correct(success_signal, error_signal,
                    patched_backend_method, expected_result)

    def test_error(self):
        """Test that the 'method' fails as expected."""
        success_signal = self.error_signal
        error_signal = self.success_signal
        patched_backend_method = self._backend_fail
        expected_result = (APP_NAME, dict(errtype='ValueError'))

        return self.assert_method_correct(success_signal, error_signal,
                    patched_backend_method, expected_result)


class SSOLoginProxyTestCase(AbstractTestCase):
    """Test the SSOLoginProxy interface."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(SSOLoginProxyTestCase, self).setUp()
        self.test_client = self.sso_client.sso_login
        self.patchable_backend = self.sso_service.sso_login.processor


class GenerateCaptchaTestCase(SSOLoginProxyTestCase):
    """Test the generate_captcha method."""

    method = 'generate_captcha'
    params = (APP_NAME, FILENAME)
    success_signal = 'CaptchaGenerated'
    error_signal = 'CaptchaGenerationError'
    backend_result = 'a captcha id'
    success_result = (APP_NAME, backend_result)


class RegisterUserTestCase(SSOLoginProxyTestCase):
    """Test the register_user method."""

    method = 'register_user'
    params = (APP_NAME, EMAIL, PASSWORD, NAME, CAPTCHA_ID, CAPTCHA_SOLUTION)
    success_signal = 'UserRegistered'
    error_signal = 'UserRegistrationError'
    backend_result = EMAIL
    success_result = (APP_NAME, backend_result)


class LoginTestCase(SSOLoginProxyTestCase):
    """Test the login method."""

    method = 'login'
    params = (APP_NAME, EMAIL, PASSWORD)
    success_signal = 'LoggedIn'
    error_signal = 'LoginError'
    backend_result = TOKEN
    success_result = (APP_NAME, EMAIL)

    @defer.inlineCallbacks
    def test_success(self):
        """Test that the 'method' works ok when the user is validated."""
        self.patch(self.patchable_backend, 'is_validated', lambda _: True)

        yield super(LoginTestCase, self).test_success()

        expected_credentials = yield self.keyring.get_credentials(APP_NAME)
        self.assertEqual(expected_credentials, TOKEN)

    def test_not_validated(self):
        """Test that the 'method' works ok when the user is not validated."""
        self.patch(self.patchable_backend, 'is_validated', lambda _: False)
        self.patch(self, 'success_signal', 'UserNotValidated')

        return super(LoginTestCase, self).test_success()

    def test_error_when_setting_credentials(self):
        """The 'error_signal' is emitted when credentials can't be set."""
        self.patch(self.patchable_backend, 'is_validated', lambda _: True)
        exc = TypeError('foo')
        self.patch(self.keyring, 'set_credentials', lambda *a: defer.fail(exc))
        patched_backend_method = lambda *a, **kw: self.backend_result
        expected_result = (APP_NAME, dict(errtype='TypeError', message='foo'))

        return self.assert_method_correct('LoginError', 'LoggedIn',
                    patched_backend_method, expected_result)


class ValidateEmailTestCase(SSOLoginProxyTestCase):
    """Test the validate_email method."""

    method = 'validate_email'
    params = (APP_NAME, EMAIL, PASSWORD, EMAIL_TOKEN)
    success_signal = 'EmailValidated'
    error_signal = 'EmailValidationError'
    backend_result = TOKEN
    success_result = (APP_NAME, EMAIL)


class RequestPasswordResetTokenTestCase(SSOLoginProxyTestCase):
    """Test the request_password_reset_token method."""

    method = 'request_password_reset_token'
    params = (APP_NAME, EMAIL)
    success_signal = 'PasswordResetTokenSent'
    error_signal = 'PasswordResetError'
    backend_result = EMAIL
    success_result = (APP_NAME, backend_result)


class SetNewPasswordTestCase(SSOLoginProxyTestCase):
    """Test the set_new_password method."""

    method = 'set_new_password'
    params = (APP_NAME, EMAIL, EMAIL_TOKEN, PASSWORD)
    success_signal = 'PasswordChanged'
    error_signal = 'PasswordChangeError'
    backend_result = EMAIL
    success_result = (APP_NAME, backend_result)


class CredentialsManagementProxyTestCase(AbstractTestCase):
    """Tests for the CredentialsManagementProxy DBus interface."""

    args = dict(foo='bar', fuh='baz')
    params = (APP_NAME, args)
    success_signal = 'CredentialsFound'
    error_signal = 'CredentialsError'
    backend_result = TOKEN
    success_result = (APP_NAME, backend_result)

    @defer.inlineCallbacks
    def setUp(self):
        yield super(CredentialsManagementProxyTestCase, self).setUp()
        self.credentials = FakedCredentials()
        self.patch(main, 'Credentials', lambda *a, **kw: self.credentials)

        self.test_client = self.sso_client.cred_manager
        self.patchable_backend = self.credentials

    def _backend_succeed(self, *args, **kwargs):
        """Make self.patchable_backend return self.backend_result."""
        return defer.succeed(self.backend_result)

    def _backend_fail(self, *args, **kwargs):
        """Make self.patchable_backend return a failed deferred."""
        return defer.fail(ValueError((args, kwargs)))


class FindCredentialsTestCase(CredentialsManagementProxyTestCase):
    """Test the find_credentials method."""

    method = 'find_credentials'

    @skipIfOS('win32', 'find_credentials_sync is only provided in Linux '
                       'due to compatibility issues with old clients.')
    @defer.inlineCallbacks
    def test_find_credentials_sync(self):
        """The credentials are asked and returned in a sync call."""
        d = defer.Deferred()

        self.test_client.call_method('find_credentials_sync',
                                APP_NAME, self.args,
                                reply_handler=d.callback,
                                error_handler=d.errback)
        creds = yield d
        self.assertEqual(creds, TOKEN)

    @skipIfOS('win32', 'find_credentials_sync is only provided in Linux '
                       'due to compatibility issues with old clients.')
    @defer.inlineCallbacks
    def test_find_credentials_sync_error(self):
        """If find_credentials_sync fails, error_handler is called."""
        self.patch(self.credentials, 'find_credentials', self._backend_fail)
        d = defer.Deferred()

        self.test_client.call_method('find_credentials_sync',
                                APP_NAME, self.args,
                                reply_handler=d.errback,
                                error_handler=d.callback)
        error = yield d
        error = error.args[0]
        self.assertEqual(error, 'ValueError')


class ClearCredentialsTestCase(CredentialsManagementProxyTestCase):
    """Test the clear_credentials method."""

    method = 'clear_credentials'
    success_signal = 'CredentialsCleared'
    success_result = (APP_NAME,)


class StoreCredentialsTestCase(CredentialsManagementProxyTestCase):
    """Test the store_credentials method."""

    method = 'store_credentials'
    success_signal = 'CredentialsStored'
    success_result = (APP_NAME,)


class RegisterTestCase(CredentialsManagementProxyTestCase):
    """Test the register method."""

    method = 'register'


class LoginOnlyTestCase(CredentialsManagementProxyTestCase):
    """Test the login method."""

    method = 'login'


class LoginEmailPasswordTestCase(CredentialsManagementProxyTestCase):
    """Test the login method."""

    method = 'login_email_password'
    backend_method = 'login'
    args = dict(email=EMAIL, password=PASSWORD)
    params = (APP_NAME, args)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Utility to spawn another program from a GLib mainloop."""

import os

# pylint: disable=E0611,F0401
from gi.repository import GLib
# pylint: enable=E0611,F0401

from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.utils import compat

logger = setup_logging("ubuntu_kylin_sso.utils.runner.glib")


NO_SUCH_FILE_OR_DIR = '[Errno 2]'


def spawn_program(args, reply_handler, error_handler):
    """Spawn the program specified by 'args' using the GLib mainloop.

    When the program finishes, 'reply_handler' will be called with a single
    argument that will be the porgram status code.

    If there is an error, error_handler will be called with an instance of
    SpawnError.

    """

    def child_watch(pid, status):
        """Handle child termination."""
        # pylint: disable=E1103
        GLib.spawn_close_pid(pid)
        # pylint: enable=E1103

        if os.WIFEXITED(status):
            status = os.WEXITSTATUS(status)
            reply_handler(status)
        else:
            msg = 'Child terminated abnormally, '\
                  'status from waitpid is %r' % status
            error_handler(msg=msg, failed_to_start=False)

    def handle_error(gerror):
        """Handle error when spawning the process."""
        failed_to_start = NO_SUCH_FILE_OR_DIR in gerror.message
        msg = 'GError is: code %r, message %r' % (gerror.code, gerror.message)
        error_handler(msg=msg, failed_to_start=failed_to_start)

    flags = GLib.SpawnFlags.DO_NOT_REAP_CHILD | \
            GLib.SpawnFlags.SEARCH_PATH | \
            GLib.SpawnFlags.STDOUT_TO_DEV_NULL | \
            GLib.SpawnFlags.STDERR_TO_DEV_NULL
    pid = None

    bytes_args = []
    for arg in args:
        if isinstance(arg, compat.text_type):
            arg = arg.encode('utf-8')
        if not isinstance(arg, compat.basestring):
            arg = compat.binary_type(arg)
        bytes_args.append(arg)

    try:
        pid, _, _, _ = GLib.spawn_async(argv=bytes_args, flags=flags)
    except GLib.GError as e:
        handle_error(e)
    else:
        logger.debug('Spawning the program %r with the glib mainloop '
                     '(returned pid is %r).', args, pid)
        GLib.child_watch_add(pid, child_watch)



# -*- coding: utf-8 -*-
#
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Platform specific constants and functions (for Qt/Twisted)."""

from twisted.internet import defer
from PyQt4.QtCore import QThread, QCoreApplication


class DeferredThread(QThread):
    """A thread that runs a given function."""

    def __init__(self, f, *args, **kwargs):
        """Initialize this thread."""
        app = QCoreApplication.instance()
        super(DeferredThread, self).__init__(app)
        self.deferred = defer.Deferred()
        self.f = f
        self.args = args
        self.kwargs = kwargs
        self.succeeded = True
        self.result = None
        self.finished.connect(self.on_finished)

    def run(self):
        """This code runs inside the thread."""
        try:
            self.result = self.f(*self.args, **self.kwargs)
        # pylint: disable=W0703
        except Exception as e:
            self.succeeded = False
            self.result = e

    def on_finished(self):
        """The thread has completed."""
        if self.succeeded:
            self.deferred.callback(self.result)
        else:
            self.deferred.errback(self.result)


# pylint: disable=C0103
def qtDeferToThread(f, *args, **kwargs):
    """A Qt-based implementation of deferToThread."""
    thread = DeferredThread(f, *args, **kwargs)
    thread.start()
    return thread.deferred



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the qt runner helper module."""

import subprocess

from PyQt4 import QtCore
from twisted.internet import defer

from ubuntu_kylin_sso.utils import compat, runner
from ubuntu_kylin_sso.utils.runner import qt
from ubuntu_kylin_sso.utils.runner.tests.test_runner import SpawnProgramTestCase


class FakedSignal(object):
    """Fake a Qt signal."""

    def __init__(self, name):
        self.name = name
        self._handlers = []

    def connect(self, handler):
        """Connect 'handler' with this signal."""
        self._handlers.append(handler)

    def emit(self, *args, **kwargs):
        """Emit this signal."""
        for handler in self._handlers:
            handler(*args, **kwargs)


class FakedProcess(object):
    """Fake a Qt Process."""

    _pid = 123456
    _error = None
    _status_code = 0

    FailedToStart = 0

    def __init__(self):
        self.pid = lambda: self._pid
        self.started = FakedSignal('started')
        self.finished = FakedSignal('finished')
        self.error = FakedSignal('error')

    def start(self, program, arguments):
        """Start this process."""
        if self._error is None:
            self.started.emit()

            args = (program,) + tuple(arguments)

            # subprocess expects bytes
            bytes_args = []
            for arg in args:
                if isinstance(arg, compat.text_type):
                    arg = arg.encode('utf-8')
                bytes_args.append(arg)

            try:
                subprocess.call(bytes_args)
            except OSError as e:
                if e.errno == 2:
                    self.error.emit(self.FailedToStart)
                else:
                    self.error.emit(e)
            except Exception as e:
                self.error.emit(e)
            else:
                self.finished.emit(self._status_code)
        else:
            self.error.emit(self._error)


class QtSpawnProgramTestCase(SpawnProgramTestCase):
    """The test suite for the spawn_program method (using Qt)."""

    use_reactor = False

    @defer.inlineCallbacks
    def setUp(self):
        yield super(QtSpawnProgramTestCase, self).setUp()
        # Since we can't mix plan qt runner and the qt4reactor, we patch
        # QProcess and fake the conditions so the qt runner is chosen
        self.process = FakedProcess()
        self.patch(QtCore, 'QProcess', lambda: self.process)
        self.patch(runner, 'is_qt4_main_loop_installed', lambda: True)

    @defer.inlineCallbacks
    def test_dont_let_it_go_success(self):
        """The QProcess instance is stored (and removed) to avoid GC."""
        # pylint: disable=W0212

        d = defer.Deferred()

        def check(status):
            """Do the check."""
            self.assertIn(self.process, qt._processes)
            d.callback(status)

        runner.qt.spawn_program(self.args,
            reply_handler=check, error_handler=lambda **kw: d.errback(kw))

        yield d

        self.assertNotIn(self.process, qt._processes)

    @defer.inlineCallbacks
    def test_dont_let_it_go_error(self):
        """The QProcess instance is stored (and removed) to avoid GC."""
        # pylint: disable=W0212

        d = defer.Deferred()

        def check(msg, failed_to_start):
            """Do the check."""
            self.assertIn(self.process, qt._processes)
            d.callback(msg)

        self.process._error = 42  # this will make the process emit 'error'

        runner.qt.spawn_program(self.args,
            reply_handler=lambda _: d.errback(AssertionError(_)),
            error_handler=check)

        yield d

        self.assertNotIn(self.process, qt._processes)



# -*- coding: utf-8 -*-
#
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the network state detection code."""

from twisted.internet.defer import inlineCallbacks

from ubuntu_kylin_sso.tests import TestCase
from ubuntu_kylin_sso.networkstate import (
    darwin,
    NetworkFailException,
)

from ubuntu_kylin_sso.networkstate.networkstates import (
    ONLINE, OFFLINE, UNKNOWN,
)

from ubuntu_kylin_sso.networkstate.darwin import (
    is_machine_connected,
    NetworkManagerState
)

from ubuntu_kylin_sso.networkstate.darwin import flags_say_reachable


REACHABLE_FLAG = 1 << 1
CONNECTION_REQUIRED_FLAG = 1 << 2


class TestSCNRFailingInDirect(TestCase):

    """Test that we handle a problem getting status in a direct call
    to check_connected_state (used by is_machine_connected) by
    raising an exception.
    """

    def test_cant_create_target(self):
        """SCNRCreateWithName returning None should cause an exception."""
        self.patch(darwin, "SCNRCreateWithName",
                   lambda _1, _2: None)
        # pylint: disable=W0212
        self.assertRaises(NetworkFailException,
                          darwin.check_connected_state)

    def test_cant_get_flags(self):
        """SCNRGetFlags returning False should cause an exception."""
        self.patch(darwin, "SCNRGetFlags",
                   lambda _1, _2: False)
        # pylint: disable=W0212
        self.assertRaises(NetworkFailException,
                          darwin.check_connected_state)


class TestFailingSCNRInCallbacks(TestCase):

    """Test that we handle a problem getting status in the separate
    listening thread by updating the status to UNKNOWN.
    """

    def expect_unknown(self, state):
        """A convenience callback that fails unless it sees UNKNOWN."""
        self.assertEquals(state, UNKNOWN)

    def test_exc_in_find_online_state(self):
        """Expect UNKNOWN from find_online_state in case of exception."""
        def fake_check_connected_state():
            "fake a broken check_connected_state"
            raise NetworkFailException()

        self.patch(darwin, "check_connected_state",
                   fake_check_connected_state)
        NetworkManagerState(self.expect_unknown)

    def test_cant_create_target(self):
        """SCNRCreateWithName returning None -> callback gets UNKNOWN."""
        self.patch(darwin, "SCNRCreateWithName", lambda _1, _2: None)
        nms = NetworkManagerState(self.expect_unknown)
        # pylint: disable=W0212
        nms._listen_on_separate_thread()

    def test_cant_set_callback(self):
        """SCNRSetCallback returning false -> callback gets UNKNOWN."""
        self.patch(darwin, "SCNRSetCallback", lambda _1, _2, _3: False)
        nms = NetworkManagerState(self.expect_unknown)
        # pylint: disable=W0212
        nms._listen_on_separate_thread()

    def test_cant_schedule_with_runloop(self):
        """SCNRScheduleWithRunLoop returning false -> callback gets UNKNOWN."""
        self.patch(darwin, "SCNRScheduleWithRunLoop",
                   lambda _1, _2, _3: False)
        nms = NetworkManagerState(self.expect_unknown)
        # pylint: disable=W0212
        nms._listen_on_separate_thread()


class TestReadingFlags(TestCase):
    """Test interpretation of flags returned from SCNR API"""

    def test_flag_reachable(self):
        """Reachable by itself is OK."""
        flag = REACHABLE_FLAG
        self.assertTrue(flags_say_reachable(flag))

    def test_flag_reachable_and_flag_connection_required(self):
        """Reachable and connection-required is NOT OK"""
        flag = REACHABLE_FLAG | CONNECTION_REQUIRED_FLAG
        self.assertFalse(flags_say_reachable(flag))

    def test_other_flagvals(self):
        """All other flag configurations are false for our purposes.

        They either indicate an iOS device, which we won't run this
        code on, or that the server we're testing for is on this
        machine or wired directly to it. These cases won't happen.
        """
        for flag in range(0, 17) + [1 << 16, 1 << 17, 1 << 18]:
            # only test cases without the reachable bit set:
            flag = flag & ~ 2
            self.assertEqual(False, flags_say_reachable(flag))


class TestNMSListeningForNWStateChanges(TestCase):
    """
    Test that the NetworkManagerState class calls the callback with
    ONLINE/OFFLINE when the state changes appropriately
    """

    @inlineCallbacks
    def setUp(self):
        """Setup array to hold state changes."""
        yield super(TestNMSListeningForNWStateChanges, self).setUp()
        self.network_changes = []

    def _listen_network_changes(self, state):
        """Fake callback function, records state changes."""
        self.network_changes.append(state)

    def test_network_state_change(self):
        """Test the changes in the network connection."""
        nms = NetworkManagerState(self._listen_network_changes)
        # pylint: disable=W0212
        nms._state_changed(2)
        nms._state_changed(0)            # 0 or anything other than 2.
        nms._state_changed(2)

        self.assertEqual(self.network_changes,
            [ONLINE, OFFLINE, ONLINE])


class TestIsMachineConnectedFunc(TestCase):
    """Simple test of is_machine_connected."""

    @inlineCallbacks
    def test_not_connected_returns_false(self):
        """test that False comes back False"""
        self.patch(darwin, "check_connected_state",
                   lambda: False)
        con = yield is_machine_connected()
        self.assertEqual(con, False)

    @inlineCallbacks
    def test_connected_returns_true(self):
        """check that True comes back True"""
        self.patch(darwin, "check_connected_state",
                   lambda: True)
        con = yield is_machine_connected()
        self.assertEqual(con, True)



# -*- coding: utf-8 -*-
#
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the Credentials module."""

import logging
import os

from twisted.internet import defer
from twisted.trial.unittest import TestCase
from ubuntuone.devtools.handlers import MementoHandler

import ubuntu_kylin_sso.main

from ubuntu_kylin_sso import credentials
from ubuntu_kylin_sso.credentials import (
    APP_NAME_KEY,
    HELP_TEXT_KEY,
    PING_URL_KEY,
    POLICY_URL_KEY,
    TC_URL_KEY,
    UI_EXECUTABLE_KEY,
    WINDOW_ID_KEY,
)
from ubuntu_kylin_sso.tests import (
    APP_NAME,
    EMAIL,
    HELP_TEXT,
    PASSWORD,
    PING_URL,
    POLICY_URL,
    TC_URL,
    TOKEN,
    WINDOW_ID,
)


# Access to a protected member of a client class
# pylint: disable=W0212

# Attribute defined outside __init__
# pylint: disable=W0201

# Instance of 'class' has no 'x' member (but some types could not be inferred)
# pylint: disable=E1103


KWARGS = {
    APP_NAME_KEY: APP_NAME,
    HELP_TEXT_KEY: HELP_TEXT,
    WINDOW_ID_KEY: WINDOW_ID,
    PING_URL_KEY: PING_URL,
    POLICY_URL_KEY: POLICY_URL,
    TC_URL_KEY: TC_URL,
    UI_EXECUTABLE_KEY: 'foo-bar-baz',
}

UI_KWARGS = {
    APP_NAME_KEY: APP_NAME,
    HELP_TEXT_KEY: HELP_TEXT,
    PING_URL_KEY: PING_URL,
    POLICY_URL_KEY: POLICY_URL,
    TC_URL_KEY: TC_URL,
    WINDOW_ID_KEY: WINDOW_ID,
}


class SampleMiscException(Exception):
    """An error to be used while testing."""


class FakedSSOLogin(object):
    """Fake a SSOLogin."""

    proxy = None

    def __init__(self, proxy):
        """Nothing."""
        FakedSSOLogin.proxy = proxy

    def login(self, *args, **kwargs):
        """Fake login."""


class BasicTestCase(TestCase):
    """Test case with a helper tracker."""

    bin_dir = 'some/bin/dir'

    @defer.inlineCallbacks
    def setUp(self):
        """Init."""
        yield super(BasicTestCase, self).setUp()
        self._called = False  # helper

        self.memento = MementoHandler()
        self.memento.setLevel(logging.DEBUG)
        credentials.logger.addHandler(self.memento)
        self.addCleanup(credentials.logger.removeHandler, self.memento)

        self.patch(credentials, 'get_bin_cmd',
                   lambda x: [os.path.join(self.bin_dir,
                                          KWARGS[UI_EXECUTABLE_KEY])])

    def _set_called(self, *args, **kwargs):
        """Set _called to True."""
        self._called = (args, kwargs)


class CredentialsTestCase(BasicTestCase):
    """Test suite for the Credentials class."""

    timeout = 5

    @defer.inlineCallbacks
    def setUp(self):
        """Init."""
        yield super(CredentialsTestCase, self).setUp()
        self.obj = credentials.Credentials(**KWARGS)


class CredentialsCallbacksTestCase(CredentialsTestCase):
    """Test suite for the Credentials callbacks."""

    def test_creation_parameters_are_stored(self):
        """Creation parameters are stored."""
        for key, value in KWARGS.items():
            self.assertEqual(getattr(self.obj, key), value)

    def test_tc_url_defaults_to_none(self):
        """The T&C url defaults to None."""
        newkw = KWARGS.copy()
        newkw.pop(TC_URL_KEY)
        self.obj = credentials.Credentials(**newkw)

        self.assertEqual(getattr(self.obj, TC_URL_KEY), None)

    def test_help_text_defaults_to_empty_string(self):
        """The T&C url defaults to the emtpy string."""
        newkw = KWARGS.copy()
        newkw.pop(HELP_TEXT_KEY)
        self.obj = credentials.Credentials(**newkw)

        self.assertEqual(getattr(self.obj, HELP_TEXT_KEY), '')

    def test_window_id_defaults_to_zero(self):
        """The T&C url defaults to 0."""
        newkw = KWARGS.copy()
        newkw.pop(WINDOW_ID_KEY)
        self.obj = credentials.Credentials(**newkw)

        self.assertEqual(getattr(self.obj, WINDOW_ID_KEY), 0)

    def test_ping_url_defaults_to_none(self):
        """The ping url defaults to None."""
        newkw = KWARGS.copy()
        newkw.pop(PING_URL_KEY)
        self.obj = credentials.Credentials(**newkw)

        self.assertEqual(getattr(self.obj, PING_URL_KEY), None)

    def test_policy_url_defaults_to_none(self):
        """The policy url defaults to None."""
        newkw = KWARGS.copy()
        newkw.pop(POLICY_URL_KEY)
        self.obj = credentials.Credentials(**newkw)

        self.assertEqual(getattr(self.obj, POLICY_URL_KEY), None)

    def test_ui_executable_default(self):
        """The ui class defaults to Qt."""
        newkw = KWARGS.copy()
        newkw.pop(UI_EXECUTABLE_KEY)
        self.obj = credentials.Credentials(**newkw)

        self.assertEqual(getattr(self.obj, UI_EXECUTABLE_KEY),
                         credentials.UI_EXECUTABLE_QT)


class FindCredentialsTestCase(CredentialsTestCase):
    """Test suite for the find_credentials method."""

    @defer.inlineCallbacks
    def test_find_credentials(self):
        """A deferred with credentials is returned when found."""
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.succeed(TOKEN))

        token = yield self.obj.find_credentials()
        self.assertEqual(token, TOKEN)

    @defer.inlineCallbacks
    def test_credentials_not_found(self):
        """find_credentials returns {} when no credentials are found."""
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.succeed(None))

        token = yield self.obj.find_credentials()
        self.assertEqual(token, {})

    @defer.inlineCallbacks
    def test_keyring_failure(self):
        """Failures from the keyring are handled."""
        expected_error = SampleMiscException()
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.fail(expected_error))

        yield self.assertFailure(self.obj.find_credentials(),
                                 SampleMiscException)


class ClearCredentialsTestCase(CredentialsTestCase):
    """Test suite for the clear_credentials method."""

    @defer.inlineCallbacks
    def test_clear_credentials(self):
        """The credentials are cleared."""
        self.patch(credentials.Keyring, 'delete_credentials',
                   lambda kr, app: defer.succeed(self._set_called(app)))

        yield self.obj.clear_credentials()
        self.assertEqual(self._called, ((APP_NAME,), {}))

    @defer.inlineCallbacks
    def test_keyring_failure(self):
        """Failures from the keyring are handled."""
        expected_error = SampleMiscException()
        self.patch(credentials.Keyring, 'delete_credentials',
                   lambda kr, app: defer.fail(expected_error))

        yield self.assertFailure(self.obj.clear_credentials(),
                                 SampleMiscException)


class StoreCredentialsTestCase(CredentialsTestCase):
    """Test suite for the store_credentials method."""

    @defer.inlineCallbacks
    def test_store_credentials(self):
        """The credentials are stored."""
        self.patch(credentials.Keyring, 'set_credentials',
                   lambda kr, *a: defer.succeed(self._set_called(*a)))

        yield self.obj.store_credentials(TOKEN)
        self.assertEqual(self._called, ((APP_NAME, TOKEN,), {}))

    @defer.inlineCallbacks
    def test_keyring_failure(self):
        """Failures from the keyring are handled."""
        expected_error = SampleMiscException()
        self.patch(credentials.Keyring, 'set_credentials',
                   lambda kr, app, token: defer.fail(expected_error))

        yield self.assertFailure(self.obj.store_credentials(TOKEN),
                                 SampleMiscException)


class RegisterTestCase(CredentialsTestCase):
    """Test suite for the register method."""

    operation = 'register'
    login_only = False
    kwargs = {}

    @defer.inlineCallbacks
    def setUp(self):
        yield super(RegisterTestCase, self).setUp()
        self.deferred = defer.Deferred()
        self.patch_inner(self.fake_inner)
        self.exe_path = os.path.join(self.bin_dir, KWARGS[UI_EXECUTABLE_KEY])
        self.inner_args = [
            self.exe_path,
            '--app_name', APP_NAME,
            '--help_text', HELP_TEXT,
            '--ping_url', PING_URL,
            '--policy_url', POLICY_URL,
            '--tc_url', TC_URL,
            '--window_id', str(WINDOW_ID),
        ]
        if self.login_only:
            self.inner_args.append('--login_only')

        self._next_inner_result = 0

        self.method_call = getattr(self.obj, self.operation)

    def patch_inner(self, f):
        """Patch the inner call."""
        self.patch(credentials.runner, 'spawn_program', f)

    def fake_inner(self, args):
        """Fake the runner.spawn_program."""
        self.deferred.callback(args)

        if self._next_inner_result == credentials.USER_SUCCESS:
            # fake that tokens were retrieved from the UI
            self.patch(credentials.Keyring, 'get_credentials',
                       lambda kr, app: defer.succeed(TOKEN))

        return defer.succeed(self._next_inner_result)

    def fail_inner(self, args):
        """Make the inner call fail."""
        return defer.fail(SampleMiscException(args))

    def assert_exc_msg_logged(self, exception_class, msg):
        """Check that 'msg' was logged as part as the logger.exception call."""
        for rec in self.memento.records:
            if rec.exc_info and rec.exc_info[0] == exception_class:
                self.assertIn(msg, rec.getMessage())
                break

    @defer.inlineCallbacks
    def test_with_existent_token(self):
        """The operation returns the credentials if already in keyring."""
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.succeed(TOKEN))

        result = yield self.method_call(**self.kwargs)

        self.assertEqual(result, TOKEN)
        self.assertFalse(self.deferred.called, 'No program was spawnned.')

    @defer.inlineCallbacks
    def test_without_existent_token_and_return_code_success(self):
        """The operation returns the credentials gathered by the inner call."""
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.succeed(None))
        self._next_inner_result = credentials.USER_SUCCESS

        result = yield self.method_call(**self.kwargs)
        self.assertEqual(result, TOKEN)

        # the ui was opened and proper params were passed
        args = yield self.deferred
        self.assertEqual(self.inner_args, args)

    @defer.inlineCallbacks
    def test_ui_executable_fall_back(self):
        """The executable falls back to the Qt UI if given does not exist."""
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.succeed(None))
        self._next_inner_result = credentials.USER_SUCCESS

        def get_bin_cmd_raises_if_not_qt(path):
            """Simulates nonexistence of exes that aren't qt."""
            if not path.endswith(credentials.UI_EXECUTABLE_QT):
                raise OSError()
            return [os.path.join(self.bin_dir,
                                credentials.UI_EXECUTABLE_QT)]

        self.patch(credentials, 'get_bin_cmd',
                   get_bin_cmd_raises_if_not_qt)

        result = yield self.method_call(**self.kwargs)
        self.assertEqual(result, TOKEN)

        # the ui was opened and proper params were passed
        args = yield self.deferred
        self.inner_args[0] = os.path.join(self.bin_dir,
                                          credentials.UI_EXECUTABLE_QT)
        self.assertEqual(self.inner_args, args)

    @defer.inlineCallbacks
    def test_raises_exception_if_no_ui_available(self):
        """If no GUI is available, raise GUINotAvailableError."""
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.succeed(None))

        def get_bin_cmd_just_raises(*args, **kwargs):
            """Workaround for lambda: raise."""
            raise OSError()

        self.patch(credentials, 'get_bin_cmd', get_bin_cmd_just_raises)

        f = self.method_call(**self.kwargs)
        yield self.assertFailure(f, credentials.GUINotAvailableError)

    @defer.inlineCallbacks
    def test_without_existent_token_and_return_code_cancel(self, exc=None):
        """The operation returns exc if defined, else UserCancellationError."""
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.succeed(None))
        self._next_inner_result = credentials.USER_CANCELLATION

        if exc is None:
            exc = credentials.UserCancellationError
        yield self.assertFailure(self.method_call(**self.kwargs), exc)

    @defer.inlineCallbacks
    def test_without_existent_token_and_return_other_code(self, result=None):
        """The operation returns CredentialsError with 'result'."""
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.succeed(None))
        self._next_inner_result = credentials.USER_CANCELLATION * 2  # other

        exc = yield self.assertFailure(self.method_call(**self.kwargs),
                                       credentials.CredentialsError)
        if result is None:
            result = self._next_inner_result
        self.assertEqual(exc.args, (result,))

    @defer.inlineCallbacks
    def test_with_exception_on_credentials(self):
        """The operation calls the error callback if a exception occurs."""
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.fail(SampleMiscException()))

        yield self.assertFailure(self.method_call(**self.kwargs),
                                 SampleMiscException)

        self.assertTrue(self.memento.check_exception(SampleMiscException))
        msg = 'Problem while getting credentials from the keyring'
        self.assert_exc_msg_logged(SampleMiscException, msg)

    @defer.inlineCallbacks
    def test_with_exception_on_inner_call(self):
        """The operation calls the error callback if a exception occurs."""
        self.patch(credentials.Keyring, 'get_credentials',
                   lambda kr, app: defer.succeed(None))
        self.patch_inner(self.fail_inner)

        yield self.assertFailure(self.method_call(**self.kwargs),
                                 SampleMiscException)

        self.assertTrue(self.memento.check_exception(SampleMiscException))
        msg = 'Problem while performing %s' % self.operation
        self.assert_exc_msg_logged(SampleMiscException, msg)


class LoginTestCase(RegisterTestCase):
    """Test suite for the login method."""

    operation = 'login'
    login_only = True


class LoginEmailPasswordTestCase(RegisterTestCase):
    """Test suite for the login_email_password method."""

    operation = 'login'
    login_only = True
    kwargs = {'email': EMAIL, 'password': PASSWORD}

    @defer.inlineCallbacks
    def setUp(self):
        yield super(LoginEmailPasswordTestCase, self).setUp()
        self.patch(ubuntu_kylin_sso.main, 'SSOLogin', FakedSSOLogin)
        self.patch_inner(self.fake_inner)
        self.inner_args = dict(app_name=APP_NAME,
                               email=EMAIL, password=PASSWORD,
                               ping_url=PING_URL)

    def patch_inner(self, f):
        """Patch the inner call."""
        self.patch(FakedSSOLogin, 'login', f)

    def fake_inner(self, **kwargs):
        """Fake the runner.spawn_program."""
        self.deferred.callback(kwargs)

        if self._next_inner_result == credentials.USER_SUCCESS:
            # fake that tokens were retrieved from the UI
            self.patch(credentials.Keyring, 'get_credentials',
                       lambda kr, app: defer.succeed(TOKEN))
            FakedSSOLogin.proxy.LoggedIn(APP_NAME, EMAIL)
        elif self._next_inner_result == credentials.USER_CANCELLATION:
            FakedSSOLogin.proxy.UserNotValidated(APP_NAME, EMAIL)
        else:
            FakedSSOLogin.proxy.LoginError(APP_NAME, {'errtype': 'foo'})

    def fail_inner(self, **kwargs):
        """Make the inner call fail."""
        raise SampleMiscException(kwargs)

    def test_ui_executable_fall_back(self):
        """This check does not apply for this test case."""

    def test_ui_executable_uses_qt_ui_if_none_available(self):
        """This check does not apply for this test case."""

    def test_raises_exception_if_no_ui_available(self):
        """This check does not apply for this test case."""

    def test_without_existent_token_and_return_code_cancel(self, exc=None):
        """The operation returns UserNotValidatedError."""
        exc = credentials.UserNotValidatedError
        return super(LoginEmailPasswordTestCase, self).\
               test_without_existent_token_and_return_code_cancel(exc=exc)

    def test_without_existent_token_and_return_other_code(self, result=None):
        """The operation returns CredentialsError with 'result'."""
        result = 'foo'
        return super(LoginEmailPasswordTestCase, self).\
               test_without_existent_token_and_return_other_code(result=result)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the Setup Account page Qt UI."""

from PyQt4 import QtGui, QtCore

from ubuntu_kylin_sso.qt import email_verification_page
from ubuntu_kylin_sso.qt.tests import (
    PageBaseTestCase,
    FakePageUiStyle,
)

# is ok to access private method/attrs in tests
# pylint: disable=W0212


class EmailVerificationTestCase(PageBaseTestCase):
    """Test the SetupAccountPage code."""

    ui_class = email_verification_page.EmailVerificationPage
    ui_signals = ('registrationSuccess',)
    ui_backend_signals = ('EmailValidated', 'EmailValidationError')

    def test_verification_code(self):
        """Test the verification value returned."""
        self.ui.ui.verification_code_edit.setText('asd123')
        value = self.ui.verification_code
        self.assertEqual(value, 'asd123')
        self.assertTrue(isinstance(value, str))

    def test_next_button(self):
        """Test the next button."""
        value = self.ui.next_button
        self.assertEqual(value, self.ui.ui.next_button)
        self.assertTrue(isinstance(value, QtGui.QPushButton))

    def test_connect_ui(self):
        """Test the connect ui method."""
        self.ui._connect_ui()
        # We expect 2 values because _connect_ui is called in the init too.
        self.assertEqual(self.ui.ui.verification_code_edit.receivers(
            QtCore.SIGNAL("textChanged(const QString&)")), 2)
        self.assertEqual(self.ui.next_button.receivers(
            QtCore.SIGNAL("clicked()")), 2)

    def test_validate_form_not_valid(self):
        """Test validate method."""
        button = FakePageUiStyle()
        self.patch(self.ui.ui, "next_button", button)
        self.ui.ui.verification_code_edit.setText('  ')
        self.ui.validate_form()
        self.assertFalse(button.isEnabled())

    def test_validate_form_valid(self):
        """Test validate method."""
        button = FakePageUiStyle()
        self.patch(self.ui.ui, "next_button", button)
        self.ui.ui.verification_code_edit.setText('asd123')
        self.ui.validate_form()
        self.assertTrue(button.isEnabled())

    def test_set_titles(self):
        """Test the set_titles method."""
        email = 'mail@example'
        self.ui.email = email
        self.ui.set_titles(email)
        self.assert_title_correct(email_verification_page.VERIFY_EMAIL_TITLE)
        expected = email_verification_page.VERIFY_EMAIL_CONTENT % \
            {"app_name": self.app_name,
             "email": email,
             }
        self.assert_subtitle_correct(expected)

    def test_initialize_page(self):
        """Test the initialization method."""
        button = FakePageUiStyle()
        self.patch(self.ui.ui, "next_button", button)
        self.ui.initializePage()
        self.assertTrue(button.properties['default'])
        self.assertFalse(button.isEnabled())
        expected = [('setButtonLayout', (([QtGui.QWizard.Stretch],), {}))]
        self.assertEqual(self.wizard.called, expected)

    def test_on_email_validation_error(self):
        """Test the validate_email method."""
        self.patch(self.ui, "show_error", self._set_called)
        error = {email_verification_page: 'error in email_verification_page'}

        self.ui.on_email_validation_error(self.app_name, error)

        expected = ((self.ui, ''), {})
        self.assertTrue(expected, self._called)
        self.assertEqual(self._overlay_hide_counter, 1)

    def test_on_email_validated(self):
        """Test the on_email_validated method."""
        email = 'email@example'
        self.ui.email = email

        self.assert_signal_emitted(self.ui.registrationSuccess, (email,),
            self.ui.on_email_validated, self.app_name, email)

    def test_validate_email(self):
        """Test the login method."""
        self.patch(self.ui, "hide_error", self._set_called)
        email = 'email@example.com'
        password = 'password'
        code = 'code'
        self.ui.email = email
        self.ui.password = password
        self.ui.ui.verification_code_edit.setText(code)

        self.ui.validate_email()

        self.assert_backend_called('validate_email',
            self.app_name, email, password, code)
        expected = ((), {})
        self.assertEqual(expected, self._called)
        self.assertEqual(self._overlay_show_counter, 1)



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Main module implementation specific for windows.

This module should never import from the multiplatform one (main/__init__.py),
but the other way around. Likewise, this module should *not* have any logic
regarding error processing or decision making about when to send a given
signal.

Also, most of the logging is being made in the main module to avoid
duplication between the different platform implementations.

"""
# pylint: disable=F0401
import win32process
import win32security
# pylint: enable=F0401

from twisted.internet import defer

from ubuntu_kylin_sso import BACKEND_EXECUTABLE
from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.utils import get_bin_cmd
from ubuntu_kylin_sso.utils.ipc import BaseClient
from ubuntu_kylin_sso.main.perspective_broker import (
    SSO_SERVICE_NAME,
    SSOLoginClient,
    CredentialsManagementClient,
    UbuntuSSOProxyBase,
)


logger = setup_logging("ubuntu_kylin_sso.main.windows")
U1_REG_PATH = r'Software\Ubuntu One'
SSO_INSTALL_PATH = 'SSOInstallPath'
SSO_BASE_PB_PORT = 50000
SSO_RESERVED_PORTS = 3000
SSO_PORT_ALLOCATION_STEP = 3  # contiguous ports for sso, u1client, and u1cp


def get_user_id():
    """Compute the user id of the currently running process."""
    process_handle = win32process.GetCurrentProcess()
    token_handle = win32security.OpenProcessToken(process_handle,
                                              win32security.TOKEN_ALL_ACCESS)
    user_sid = win32security.GetTokenInformation(token_handle,
                                              win32security.TokenUser)[0]
    sid_parts = str(user_sid).split("-")
    return int(sid_parts[-1])


def get_sso_pb_port():
    """Compute the port the SSO service should run on per-user."""
    uid_modulo = get_user_id() % SSO_RESERVED_PORTS
    return SSO_BASE_PB_PORT + uid_modulo * SSO_PORT_ALLOCATION_STEP


class DescriptionFactory(object):
    """Factory that provides the server and client descriptions."""

    client_description_pattern = 'tcp:host=127.0.0.1:port=%s'
    server_description_pattern = 'tcp:%s:interface=127.0.0.1'

    def __init__(self):
        """Create a new instance."""
        self.port = get_sso_pb_port()
        self.server = self.server_description_pattern % self.port
        self.client = self.client_description_pattern % self.port


class UbuntuSSOProxy(UbuntuSSOProxyBase):
    """Object that exposes the diff referenceable objects."""

    name = SSO_SERVICE_NAME

    @property
    def description(self):
        """Get the description on which the SSO pb is running."""
        return DescriptionFactory()

    @property
    def cmdline(self):
        """Get the command line to activate an executable."""
        return get_bin_cmd(BACKEND_EXECUTABLE)


class UbuntuSSOClient(BaseClient):
    """Base client that provides remote access to the sso API."""

    name = SSO_SERVICE_NAME

    clients = {
        'sso_login': SSOLoginClient,
        'cred_manager': CredentialsManagementClient,
    }

    service_name = UbuntuSSOProxy.name
    service_description = UbuntuSSOProxy.description
    service_cmdline = UbuntuSSOProxy.cmdline


@defer.inlineCallbacks
def get_sso_client():
    """Get a client to access the SSO service."""
    result = UbuntuSSOClient()
    yield result.connect()
    defer.returnValue(result)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Network state detection on OS X.

is_machine_connected(): (deferred) returns connected state as bool
NetworkManagerState: class with listening thread, calls back with state changes
"""

from twisted.internet import defer

from threading import Thread

from ubuntu_kylin_sso.networkstate import NetworkFailException
from ubuntu_kylin_sso.networkstate.networkstates import (ONLINE, OFFLINE, UNKNOWN)
from ubuntu_kylin_sso.logger import setup_logging
logger = setup_logging("ubuntu_kylin_sso.networkstate")

HOSTNAME_TO_CHECK = 'login.ubuntu.com'

from ctypes import (
    CDLL,
    POINTER,
    CFUNCTYPE,
    Structure,
    pointer,
    c_bool,
    c_long,
    c_void_p,
    c_uint32)

from ctypes.util import find_library

# pylint: disable=C0103

# Functions and constants below are from
# /System/Library/CoreFoundation.framework/
CoreFoundationPath = find_library("CoreFoundation")
CoreFoundation = CDLL(CoreFoundationPath)

# CFRunLoopRef CFRunLoopGetCurrent()
CFRunLoopGetCurrent = CoreFoundation.CFRunLoopGetCurrent
CFRunLoopGetCurrent.restype = c_void_p
CFRunLoopGetCurrent.argtypes = []

# void CFRelease(CFTypeRef)
CFRelease = CoreFoundation.CFRelease
CFRelease.restype = None
CFRelease.argtypes = [c_void_p]

# void CFRunLoopRun()
CFRunLoopRun = CoreFoundation.CFRunLoopRun

# const CFStringRef kCFRunLoopDefaultMode
# pylint: disable=E1101
kCFRunLoopDefaultMode = c_void_p.in_dll(CoreFoundation,
                                        "kCFRunLoopDefaultMode")


# Functions and constants below are from
# /System/Library/SystemConfiguration.framework/
SystemConfigurationPath = find_library("SystemConfiguration")

# SystemConfiguration abbreviated as "SC" below:
SC = CDLL(SystemConfigurationPath)

# "SCNetworkReachability" functions abbreviated to "SCNR*" here.

# SCNetworkReachabilityRef
# SCNetworkReachabilityCreateWithName(CFAllocatorRef, const char *)
SCNRCreateWithName = SC.SCNetworkReachabilityCreateWithName
SCNRCreateWithName.restype = c_void_p

# Boolean SCNetworkReachabilityGetFlags(SCNetworkReachabilityRef,
#                                       SCNetworkReachabilityFlags)
SCNRGetFlags = SC.SCNetworkReachabilityGetFlags
SCNRGetFlags.restype = c_bool
SCNRGetFlags.argtypes = [c_void_p,
                         POINTER(c_uint32)]

SCNRScheduleWithRunLoop = SC.SCNetworkReachabilityScheduleWithRunLoop
SCNRScheduleWithRunLoop.restype = c_bool
SCNRScheduleWithRunLoop.argtypes = [c_void_p,
                                    c_void_p,
                                    c_void_p]

# ctypes callback type to match SCNetworkReachabilityCallback
# void (*SCNetworkReachabilityCallback) (SCNetworkReachabilityRef,
#                                        SCNetworkReachabilityFlags,
#                                        void *)
SCNRCallbackType = CFUNCTYPE(None, c_void_p, c_uint32, c_void_p)
# NOTE: need to keep this reference alive as long as a callback might occur.

# Boolean SCNetworkReachabilitySetCallback(SCNetworkReachabilityRef,
#                                          SCNetworkReachabilityCallback,
#                                          SCNetworkReachabilityContext)
SCNRSetCallback = SC.SCNetworkReachabilitySetCallback
SCNRSetCallback.restype = c_bool
SCNRSetCallback.argtypes = [c_void_p,
                            SCNRCallbackType,
                            c_void_p]
# pylint: enable=E1101


def check_connected_state():
    """Calls Synchronous SCNR API, returns bool."""
    target = SCNRCreateWithName(None, HOSTNAME_TO_CHECK)
    if target is None:
        logger.error("Error creating network reachability reference.")
        raise NetworkFailException()

    flags = c_uint32(0)
    ok = SCNRGetFlags(target, pointer(flags))
    CFRelease(target)

    if not ok:
        logger.error("Error getting reachability status of '%s'" %
                     HOSTNAME_TO_CHECK)
        raise NetworkFailException()

    return flags_say_reachable(flags.value)


def flags_say_reachable(flags):
    """Check flags returned from SCNetworkReachability API. Returns bool.

    Requires some logic:
    reachable_flag isn't enough on its own.

    A down wifi will return flags = 7, or reachable_flag and
    connection_required_flag, meaning that the host *would be*
    reachable, but you need a connection first.  (And then you'd
    presumably be best off checking again.)
    """
    # values from SCNetworkReachability.h
    reachable_flag = 1 << 1
    connection_required_flag = 1 << 2

    if flags & connection_required_flag:
        return False
    elif flags & reachable_flag:
        return True
    else:
        return False


class SCNRContext(Structure):

    """A struct to send as SCNetworkReachabilityContext to SCNRSetCallback.

    We don't use the fields currently.
    """

    _fields_ = [("version", c_long),
                ("info", c_void_p),
                ("retain", c_void_p),           # func ptr
                ("release", c_void_p),          # func ptr
                ("copyDescription", c_void_p)]  # func ptr


class NetworkManagerState(object):

    """Probe Network State and receive callbacks on changes.

    This class uses both synchronous and async API from the
    SystemConfiguration framework.

    To use: Initialize with a callback function, then call
    find_online_state. The callback will be called once immediately
    with the current state and then only on state changes.

    Any exceptions in checking state will result in the callback being
    called with UNKNOWN. At this point the listening thread is no
    longer runing, and a new NetworkManagerState should be created.

    NOTE: the callback will be called from the separate listening
    thread, except for the first call.
    """

    def __init__(self, result_cb):
        """Initialize and save result callback function.

        result_cb should take one argument, a networkstate object.

        The callback will be called with one of ONLINE, OFFLINE, or
        UNKNOWN, as defined in networkstates.py.
        """
        self.result_cb = result_cb
        self.listener_thread = None

    def _state_changed(self, flags):
        """Testable callback called by reachability_state_changed_cb.

        Used because reachability_state_changed_cb has to have a
        particular method signature.

        Clients should not call this method.
        """
        if flags_say_reachable(flags):
            self.result_cb(ONLINE)
        else:
            self.result_cb(OFFLINE)

    def _listen_on_separate_thread(self):
        """In separate thread, setup callback and listen for changes.

        On error, calls result_cb(UNKNOWN) and returns.
        """

        def reachability_state_changed_cb(targetref, flags, info):
            """Callback for SCNetworkReachability API

            This callback is passed to the SCNetworkReachability API,
            so its method signature has to be exactly this. Therefore,
            we declare it here and just call _state_changed with
            flags."""
            self._state_changed(flags)

        c_callback = SCNRCallbackType(reachability_state_changed_cb)
        context = SCNRContext(0, None, None, None, None)

        target = SCNRCreateWithName(None, HOSTNAME_TO_CHECK)
        if target is None:
            logger.error("Error creating SCNetworkReachability target")
            self.result_cb(UNKNOWN)
            return

        ok = SCNRSetCallback(target, c_callback, pointer(context))
        if not ok:
            logger.error("error setting SCNetworkReachability callback")
            CFRelease(target)
            self.result_cb(UNKNOWN)
            return

        ok = SCNRScheduleWithRunLoop(target,
                                     CFRunLoopGetCurrent(),
                                     kCFRunLoopDefaultMode)
        if not ok:
            logger.error("error scheduling on runloop: SCNetworkReachability")
            CFRelease(target)
            self.result_cb(UNKNOWN)
            return

        CFRunLoopRun()

        CFRelease(target)               # won't happen

    def _start_listening_thread(self):
        """Start the separate listener thread.

        Currently will not start one more than once.
        Should be OK because we don't expect errors the listen method.

        To add more error handling support, you could either add a
        call to join and re-start, or client could just create a new
        NetworkManagerState.
        """

        if self.listener_thread is None:
            self.listener_thread = Thread(
                target=self._listen_on_separate_thread,
                name="Ubuntu SSO Network Connection Monitor")
            self.listener_thread.daemon = True
            self.listener_thread.start()

    def find_online_state(self):
        """Calls callback with current state. Starts listening thread."""
        try:
            if check_connected_state():
                self.result_cb(ONLINE)
            else:
                self.result_cb(OFFLINE)

        except Exception:               # pylint: disable=W0703
            logger.exception("Getting state from SCNetworkReachability")
            self.result_cb(UNKNOWN)
            return                      # don't start thread on error

        self._start_listening_thread()


def is_machine_connected():
    """Return a deferred that when fired, returns online state as a bool.

    Raises NetworkFailException for errors.
    """
    try:
        return defer.succeed(check_connected_state())
    except Exception as e:                # pylint: disable=W0703
        logger.exception("Exception calling check_connected_state:")
        return defer.fail(NetworkFailException(e))



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the parse_args function."""

from __future__ import unicode_literals

import sys

from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso.utils import ui as main


APP_NAME = 'Foo Bar'
DEFAULTS = dict(
    app_name=APP_NAME,
    help_text='',
    login_only=False,
    ping_url='',
    policy_url='',
    tc_url='',
    window_id=0,
)
DUMMY_TEXT = """Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed
risus orci, lacinia ac tincidunt fermentum, vestibulum suscipit orci. Aliquam
quis aliquet magna. Morbi vitae ligula ut libero porttitor imperdiet.
Vestibulum et ipsum sapien, pellentesque ultricies risus. Aenean in lectus
orci. Cras a lorem sollicitudin dui mollis varius. In hac habitasse platea
dictumst."""
PROGRAM = 'foo-bar'
SOME_URL = 'http://example.com/foo/bar'


class ParseArgsTestCase(TestCase):
    """Test case for the parse_args function."""

    def assert_parse_args_custom_field(self, option_name=None,
                                       option_value=None):
        """The args are correctly set when using a custom value."""
        expected = dict(DEFAULTS)
        argv = [PROGRAM, '--app_name=%s' % APP_NAME]

        if option_name is not None:
            if option_value is not None:
                argv.append('--%s=%s' % (option_name, option_value))
                expected[option_name] = option_value
            else:
                argv.append('--%s' % (option_name,))
                expected[option_name] = True

        self.patch(sys, 'argv', argv)

        result = main.parse_args()

        # Access to a protected member, pylint: disable=W0212
        self.assertEqual(expected, dict(result._get_kwargs()))

    def test_parse_args_defaults(self):
        """The args are correctly set when using defaults."""
        self.assert_parse_args_custom_field()

    def test_parse_args_help_text(self):
        """The args are correctly set when using a custom help_text."""
        self.assert_parse_args_custom_field('help_text', DUMMY_TEXT)

    def test_parse_args_ping_url(self):
        """The args are correctly set when using a custom ping_url."""
        self.assert_parse_args_custom_field('ping_url', SOME_URL)

    def test_parse_args_policy_url(self):
        """The args are correctly set when using a custom policy_url."""
        self.assert_parse_args_custom_field('policy_url', SOME_URL)

    def test_parse_args_tc_url(self):
        """The args are correctly set when using a custom tc_url."""
        self.assert_parse_args_custom_field('tc_url', SOME_URL)

    def test_parse_args_window_id(self):
        """The args are correctly set when using a custom window_id."""
        self.assert_parse_args_custom_field('window_id', 42)

    def test_parse_args_login_only(self):
        """The args are correctly set when using a custom login_only."""
        self.assert_parse_args_custom_field('login_only')



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the Setup Account page Qt UI."""

from PyQt4 import QtGui, QtCore

from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso.tests import CAPTCHA_PATH

from ubuntu_kylin_sso.qt import common, setup_account_page as gui
from ubuntu_kylin_sso.qt.tests import PageBaseTestCase


class SetupAccountTestCase(PageBaseTestCase):
    """Test the SetupAccountPage code."""

    ui_class = gui.SetupAccountPage
    ui_signals = ('userRegistered', 'signIn')
    ui_backend_signals = ('CaptchaGenerated', 'CaptchaGenerationError',
                          'UserRegistered', 'UserRegistrationError')

    def test_password_receiver(self):
        """Check the initial state of SetupAccount."""
        self.assertEqual(self.ui.ui.password_edit.receivers(
            QtCore.SIGNAL('textEdited(QString)')), 2)

    def test_empty_name(self):
        """Status when the name field is empty.

        * Name Assistance label visible.
        * Correct error message in it.
        """
        self.ui.ui.name_edit.setText("")
        self.ui.name_assistance()
        self.assertTrue(self.ui.ui.name_assistance.isVisible())
        self.assert_error_correct(self.ui.ui.name_assistance, gui.EMPTY_NAME,
                                  max_width=self.ui.header.max_title_width)

    def test_hide_error_on_refresh_clicked(self):
        """Hide form errors when the user click to refresh the captcha."""
        self.ui.show_error('error')
        self.assert_error_correct(self.ui.form_errors_label, 'error',
            max_width=self.ui.header.max_title_width)

        self.ui.ui.refresh_label.linkActivated.emit('error')

        message = compat.text_type(self.ui.form_errors_label.text())
        self.assertEqual(message, ' ')

    def test_enable_setup_button_with_visible_check(self):
        """Test _enable_setup_button method with terms check visible."""
        self.patch(self.ui, "set_up_button", QtGui.QPushButton())

        self.ui.ui.name_edit.setText('name')
        email = 'email@example.com'
        self.ui.ui.email_edit.setText(email)
        self.ui.ui.confirm_email_edit.setText(email)
        password = 'T3st3rqwe'
        self.ui.ui.password_edit.setText(password)
        self.ui.ui.confirm_password_edit.setText(password)
        self.ui.ui.captcha_solution_edit.setText('captcha solution')
        self.ui.terms_checkbox.setChecked(True)

        self.ui.terms_checkbox.setVisible(True)
        self.ui.ui.captcha_solution_edit.textEdited.emit('')
        self.assertTrue(self.ui.set_up_button.isEnabled())

    def test_enable_setup_button_without_visible_check(self):
        """Test _enable_setup_button method with terms check not visible."""
        self.patch(self.ui, "set_up_button", QtGui.QPushButton())

        self.ui.ui.name_edit.setText('name')
        email = 'email@example.com'
        self.ui.ui.email_edit.setText(email)
        self.ui.ui.confirm_email_edit.setText(email)
        password = 'T3st3rqwe'
        self.ui.ui.password_edit.setText(password)
        self.ui.ui.confirm_password_edit.setText(password)
        self.ui.ui.captcha_solution_edit.setText('captcha solution')

        self.ui.terms_checkbox.setVisible(False)
        self.ui.ui.captcha_solution_edit.textEdited.emit('')
        self.assertTrue(self.ui.set_up_button.isEnabled())

    def test_password_default_assistance(self):
        """Status when the password line edit receive focus and shows popup.

        * Password assistance contains the right message.
        """
        self.ui.ui.password_edit.setText("")
        self.ui.ui.confirm_password_edit.setText("")
        common.password_default_assistance(self.ui.ui.password_assistance)
        self.ui.focus_changed(None, self.ui.ui.password_edit)
        self.assertIn(
            common.NORMAL % common.PASSWORD_LENGTH,
            compat.text_type(self.ui.ui.password_assistance.text()),
        )
        self.assertIn(
            common.NORMAL % common.PASSWORD_UPPER,
            compat.text_type(self.ui.ui.password_assistance.text()),
        )
        self.assertIn(
            common.NORMAL % common.PASSWORD_DIGIT,
            compat.text_type(self.ui.ui.password_assistance.text()),
        )

    def test_password_focus_gain(self):
        """Check functions execution when focus_changed() is executed."""
        self.ui.ui.password_assistance.setVisible(False)
        self.assertFalse(self.ui.ui.password_assistance.isVisible())
        self.patch(self.ui, 'name_assistance', self._set_called)
        self.ui.focus_changed(None, self.ui.ui.password_edit)
        self.assertTrue(self.ui.ui.password_assistance.isVisible())

    def test_name_focus_lost(self):
        """Check functions execution when focus_changed() is executed."""
        self.patch(self.ui, 'name_assistance', self._set_called)
        self.ui.focus_changed(self.ui.ui.name_edit, None)
        self.assertTrue(self._called)

    def test_email_focus_lost(self):
        """Check functions execution when focus_changed() is executed."""
        self.patch(self.ui, 'email_assistance', self._set_called)
        self.ui.focus_changed(self.ui.ui.email_edit, None)
        self.assertTrue(self._called)

    def test_confirm_email_focus_lost(self):
        """Check functions execution when focus_changed() is executed."""
        self.patch(self.ui, 'confirm_email_assistance', self._set_called)
        self.ui.focus_changed(self.ui.ui.confirm_email_edit, None)
        self.assertTrue(self._called)

    def test_confirm_password_focus_lost(self):
        """Check functions execution when focus_changed() is executed."""
        self.patch(common, 'password_check_match', self._set_called)
        self.ui.focus_changed(self.ui.ui.confirm_password_edit, None)
        self.assertTrue(self._called)

    def test_on_user_registered(self):
        """Test on_user_registered method."""
        self.patch(self.ui, "validate_form", lambda: True)
        self.patch(self.ui, "hide_error", self._set_called)

        email = 'email@example'
        password = 'password'
        name = 'name'
        captcha_id = 1
        captcha_solution = 'solution'
        self.ui.ui.email_edit.setText(email)
        self.ui.ui.password_edit.setText(password)
        self.ui.ui.name_edit.setText(name)
        self.ui.captcha_id = captcha_id
        self.ui.ui.captcha_solution_edit.setText(captcha_solution)

        self.ui.set_next_validation()

        self.assert_backend_called('register_user',
            self.app_name, email, password, name, captcha_id,
            captcha_solution)
        expected = ((), {})
        self.assertEqual(expected, self._called)

    def test_set_next_validation(self):
        """Test on_user_registered method."""
        email = 'email@example'
        self.ui.ui.email_edit.setText(email)
        self.assert_signal_emitted(self.ui.userRegistered, (email,),
            self.ui.on_user_registered, self.app_name, email)

    def test_captcha_image_is_requested_as_startup(self):
        """The captcha image is requested at startup."""
        # assert generate_captcha was called
        self.assert_backend_called('generate_captcha',
            self.app_name, self.ui.captcha_file)

    def test_set_translated_strings(self):
        """Test the translated string method."""
        expected = gui.REGISTER_TITLE.format(app_name=self.app_name)
        self.assert_title_correct(expected)
        self.assert_subtitle_correct(self.ui.help_text)

    def test_initialize_page(self):
        """Widgets are properly initialized."""
        self.ui.initializePage()

        # set up account button
        expected = [QtGui.QWizard.Stretch, QtGui.QWizard.CustomButton3]
        self.assertIn(('setButtonLayout', ((expected,), {})),
                       self.ui.wizard().called)
        self.assertEqual(gui.SET_UP_ACCOUNT_BUTTON,
                         self.ui.buttonText(QtGui.QWizard.CustomButton3))
        self.assertTrue(self.ui.wizard().button(QtGui.QWizard.CustomButton3) is
                        self.ui.set_up_button)
        self.assertFalse(self.ui.set_up_button.isEnabled())

        self.assertFalse(self.ui.captcha_received)

        # labels
        self.assertEqual(compat.text_type(self.ui.ui.name_label.text()),
            gui.NAME_ENTRY)
        self.assertEqual(compat.text_type(self.ui.ui.email_label.text()),
            gui.EMAIL)
        self.assertEqual(
                compat.text_type(self.ui.ui.confirm_email_label.text()),
                gui.RETYPE_EMAIL)
        self.assertEqual(compat.text_type(self.ui.ui.password_label.text()),
            gui.PASSWORD)
        self.assertEqual(
                compat.text_type(self.ui.ui.confirm_password_label.text()),
                gui.RETYPE_PASSWORD)
        expected = gui.LINK_STYLE.format(link_url='#',
            link_text=gui.SIGN_IN_LABEL)
        self.assertEqual(
                compat.text_type(self.ui.ui.sign_in_label.text()),
                expected)

        # connections
        self.assertEqual(self.ui.set_up_button.receivers(
            QtCore.SIGNAL("clicked(const bool)")), 1)

        # assistants
        self.assertFalse(self.ui.ui.name_assistance.isVisible())
        self.assertFalse(self.ui.ui.email_assistance.isVisible())
        self.assertFalse(self.ui.ui.confirm_email_assistance.isVisible())
        self.assertFalse(self.ui.ui.password_assistance.isVisible())
        self.assertTrue(self.ui.ui.refresh_label.isVisible())

    def test_connect_ui(self):
        """Test the connect ui method."""
        self.assertEqual(self.ui.ui.password_edit.receivers(
            QtCore.SIGNAL("textEdited(const QString&)")), 2)
        self.assertEqual(self.ui.ui.refresh_label.receivers(
            QtCore.SIGNAL("linkActivated(const QString&)")), 2)
        self.assertEqual(self.ui.ui.name_edit.receivers(
            QtCore.SIGNAL("textEdited(const QString&)")), 1)
        self.assertEqual(self.ui.ui.email_edit.receivers(
            QtCore.SIGNAL("textEdited(const QString&)")), 1)
        self.assertEqual(self.ui.ui.confirm_email_edit.receivers(
            QtCore.SIGNAL("textEdited(const QString&)")), 1)
        self.assertEqual(self.ui.ui.password_edit.receivers(
            QtCore.SIGNAL("textEdited(const QString&)")), 2)
        self.assertEqual(self.ui.ui.confirm_password_edit.receivers(
            QtCore.SIGNAL("textEdited(const QString&)")), 1)
        self.assertEqual(self.ui.ui.sign_in_label.receivers(
            QtCore.SIGNAL("linkActivated(const QString&)")), 1)

    def test_connect_ui_2(self):
        """Test more connections from connect_ui"""
        self.assertEqual(self.ui.ui.captcha_solution_edit.receivers(
            QtCore.SIGNAL("textEdited(const QString&)")), 1)
        self.assertEqual(self.ui.terms_checkbox.receivers(
            QtCore.SIGNAL("stateChanged(bool)")), 1)

    test_connect_ui_2.skip = "These oscilate 0/1 (LP:1046885)"

    def test_set_up_button_clicked(self):
        """Validation and overlay is shown when the setupbutton is clicked."""
        self.patch(self.ui, 'set_next_validation', self._set_called)
        self.ui.initializePage()
        self.ui.captcha_received = True

        self.ui.set_up_button.clicked.emit(False)
        self.assertEqual(self._called, ((False,), {}))

    def test_set_error_message(self):
        """Check the state of the label after calling: set_error_message."""
        self.ui.email_assistance()
        self.ui.set_error_message(self.ui.ui.email_assistance, "message")
        self.assertTrue(self.ui.ui.email_assistance.isVisible())
        self.assert_error_correct(self.ui.ui.email_assistance, "message",
                                  max_width=self.ui.header.max_title_width)

    def test_blank_name(self):
        """Status when the name field is blank (spaces).

        * Name Assistance label visible.
        * Correct error message in it.
        """
        self.ui.ui.name_edit.setText("  ")
        self.ui.name_assistance()
        self.assertTrue(self.ui.ui.name_assistance.isVisible())
        self.assert_error_correct(self.ui.ui.name_assistance, gui.EMPTY_NAME,
                                  max_width=self.ui.header.max_title_width)

    def test_valid_name(self):
        """Status when the name field is valid.

        * Name Assistance label invisible.
        """
        self.ui.ui.name_edit.setText("John Doe")
        self.ui.name_assistance()
        self.assertFalse(self.ui.ui.name_assistance.isVisible())

    def test_invalid_email(self):
        """Status when the email field has no @.

        * Email Assistance label visible.
        * Correct error message in it.
        """
        self.ui.ui.email_edit.setText("foobar")
        self.ui.email_assistance()
        self.assertTrue(self.ui.ui.email_assistance.isVisible())
        self.assert_error_correct(self.ui.ui.email_assistance,
                                  gui.INVALID_EMAIL,
                                  max_width=self.ui.header.max_title_width)

    def test_valid_email(self):
        """Status when the email field has a @.

        * Email Assistance label invisible.
        """
        self.ui.ui.email_edit.setText("foo@bar")
        self.ui.email_assistance()
        self.assertFalse(self.ui.ui.email_assistance.isVisible())

    def test_matching_emails(self):
        """Status when the email fields match.

        * Email Assistance label invisible.
        """
        self.ui.ui.email_edit.setText("foo@bar")
        self.ui.ui.confirm_email_edit.setText("foo@bar")
        self.ui.confirm_email_assistance()
        self.assertFalse(self.ui.ui.confirm_email_assistance.isVisible())

    def test_not_matching_emails(self):
        """Status when the email fields don't match.

        * Email Assistance label visible.
        * Correct error message.
        """
        self.ui.ui.email_edit.setText("foo@bar")
        self.ui.ui.confirm_email_edit.setText("foo@baz")
        self.ui.confirm_email_assistance()
        self.assertTrue(self.ui.ui.confirm_email_assistance.isVisible())
        self.assert_error_correct(self.ui.ui.confirm_email_assistance,
                                  gui.EMAIL_MATCH,
                                  max_width=self.ui.header.max_title_width)

    def test_focus_changed_password_visibility(self):
        """Check visibility changes when focus_changed() is executed."""
        self.ui.focus_changed(None, self.ui.ui.password_edit)
        self.assertTrue(self.ui.ui.password_assistance.isVisible())

    def test_hide_event(self):
        """Check that hide_event works when set_up_button is none."""
        self.assertEqual(self.ui.set_up_button, None)
        self.ui.hideEvent(QtGui.QHideEvent())

    def test_show_event(self):
        """Check that show_event works when set_up_button is none."""
        self.assertEqual(self.ui.set_up_button, None)
        self.ui.show()
        self.addCleanup(self.ui.hide)

    def test_on_captcha_generated(self):
        """Check that the captcha image is handled correctly."""

        self.ui.captcha_file = CAPTCHA_PATH
        self.ui.on_captcha_generated(u"fake app name",
                                     "fake id")
        pixmap_empty = self.ui.captcha_image.size().isEmpty()
        self.assertEqual(pixmap_empty, False)

    def test_on_captcha_refreshing_visible(self):
        """Check the state of the overlay on captcha refreshing."""
        self.ui.hide_overlay()

        self.assertEqual(self._overlay_show_counter, 0)
        self.assertTrue(self.ui.isEnabled())

        self.ui.on_captcha_refreshing()

        self.assertFalse(self.ui.isEnabled())
        self.assertEqual(self._overlay_show_counter, 1)

    def test_on_captcha_refreshing_not_visible(self):
        """Check the state of the overlay on captcha refreshing."""
        self.ui.hide_overlay()

        self.assertEqual(self._overlay_show_counter, 0)
        self.assertTrue(self.ui.isEnabled())

        self.ui.hide()
        self.ui.on_captcha_refreshing()

        self.assertEqual(self._overlay_show_counter, 0)
        self.assertTrue(self.ui.isEnabled())

    def test_on_captcha_refresh_complete(self):
        """Check the state of the overlay on captcha refreshing complete."""
        self.assertEqual(self._overlay_hide_counter, 0)

        self.ui.on_captcha_refresh_complete()

        self.assertEqual(self._overlay_hide_counter, 1)

    def test_hide_error_on_refresh_captcha(self):
        """Test that the errors are hidden on refresh captcha."""
        self.ui.show_error('error-message')
        self.ui.ui.refresh_label.linkActivated.emit('link')
        self.assertEqual(self.ui.form_errors_label.text(), ' ')



# -*- coding: utf-8 -*-
#
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the network state detection code."""

from collections import defaultdict

from twisted.internet.defer import inlineCallbacks
from mocker import ARGS, KWARGS, ANY, MockerTestCase

from ubuntu_kylin_sso.tests import TestCase
from ubuntu_kylin_sso.networkstate import (
    linux,
    NetworkFailException,
    NetworkManagerState,
)

from ubuntu_kylin_sso.networkstate.networkstates import (
    ONLINE, OFFLINE, UNKNOWN,
    NM_STATE_ASLEEP,
    NM_STATE_ASLEEP_OLD,
    NM_STATE_CONNECTING,
    NM_STATE_CONNECTING_OLD,
    NM_STATE_CONNECTED_OLD,
    NM_STATE_CONNECTED_LOCAL,
    NM_STATE_CONNECTED_SITE,
    NM_STATE_CONNECTED_GLOBAL,
    NM_STATE_DISCONNECTED,
    NM_STATE_DISCONNECTED_OLD,
    NM_STATE_UNKNOWN,
)

from ubuntu_kylin_sso.networkstate.linux import (
    is_machine_connected,
    NM_DBUS_INTERFACE,
    NM_DBUS_OBJECTPATH,
)


class TestException(Exception):
    """An exception to test error conditions."""
    def get_dbus_name(self):
        """A fake dbus name for this exception."""
        return "Test Exception Message"


class FakeNetworkManagerState(object):

    """Fake Network Manager State."""

    connection_state = None

    def __init__(self, function):
        """Initialize Fake class."""
        self.call_function = function

    def find_online_state(self):
        """Fake find_online_state for linux module."""
        self.call_function(self.connection_state)


class FakeDBusMatch(object):

    """Fake a DBus match."""

    def __init__(self, name, callback, interface):
        self.name = name
        self.callback = callback
        self.interface = interface
        self.removed = False

    def remove(self):
        """Stop calling the handler function on remove."""
        self.removed = True


class FakeDBusInterface(object):

    """Fake DBus Interface."""

    def __init__(self):
        self._signals = defaultdict(list)

    # pylint: disable=C0103
    def Get(self, *args, **kwargs):
        """Fake Get."""
    # pylint: enable=C0103

    def connect_to_signal(self, signal_name, handler_function, dbus_interface):
        """Fake connect_to_signal."""
        match = FakeDBusMatch(signal_name, handler_function, dbus_interface)
        self._signals[signal_name].append(match)
        return match

    def emit_signal(self, signal_name, state):
        """Emit signal for network state change."""
        for match in self._signals[signal_name]:
            match.callback(state)


class FakeSystemBus(object):

    """Fake SystemBus."""

    objects = {(NM_DBUS_INTERFACE, NM_DBUS_OBJECTPATH, True): object()}

    def get_object(self, interface, object_path, follow_name_owner_changes):
        """Fake get_object."""
        key = (interface, object_path, follow_name_owner_changes)
        return self.objects[key]


class TestConnection(TestCase):

    """Test the state of the connection.

    This TestCase tests over all the connection states possible.

    """

    @inlineCallbacks
    def setUp(self):
        """Setup the mocker dbus object tree."""
        yield super(TestConnection, self).setUp()
        self.patch(linux, "NetworkManagerState", FakeNetworkManagerState)
        self.patch(linux.dbus, 'SystemBus', FakeSystemBus)
        self.nm_interface = FakeDBusInterface()
        self.patch(linux.dbus, 'Interface', lambda *a: self.nm_interface)
        self.network_changes = []

    def _listen_network_changes(self, state):
        """Fake callback function."""
        self.network_changes.append(state)

    def test_network_state_change(self):
        """Test the changes in the network connection."""
        nms = NetworkManagerState(self._listen_network_changes)

        nms.find_online_state()

        self.nm_interface.emit_signal('StateChanged',
            NM_STATE_CONNECTED_GLOBAL)
        self.nm_interface.emit_signal('StateChanged', NM_STATE_DISCONNECTED)
        self.nm_interface.emit_signal('StateChanged',
            NM_STATE_CONNECTED_GLOBAL)

        self.assertEqual(nms.state_signal.name, "StateChanged")
        self.assertEqual(nms.state_signal.callback, nms.state_changed)
        self.assertEqual(nms.state_signal.interface,
            "org.freedesktop.NetworkManager")
        self.assertEqual(self.network_changes,
            [ONLINE, ONLINE, OFFLINE, ONLINE])
        self.assertFalse(nms.state_signal.removed)

    @inlineCallbacks
    def test_is_machine_connected_nm_state_online(self):
        """Callback given ONLINE should mean we are online"""
        self.patch(FakeNetworkManagerState, "connection_state",
                   ONLINE)
        d = yield is_machine_connected()
        self.assertTrue(d)

    @inlineCallbacks
    def test_is_machine_connected_nm_state_offline(self):
        """Callback given OFFLINE should mean we are offline"""
        self.patch(FakeNetworkManagerState, "connection_state",
                   OFFLINE)
        d = yield is_machine_connected()
        self.assertFalse(d)

    @inlineCallbacks
    def test_is_machine_connected_nm_state_unknown(self):
        """Callback given ONLINE should mean we are not online"""
        self.patch(FakeNetworkManagerState, "connection_state",
                   UNKNOWN)
        d = yield is_machine_connected()
        self.assertFalse(d)

    @inlineCallbacks
    def test_is_machine_connected_callback_error(self):
        """Test bad argument to is_machine_connected's internal callback.

        Passing anything other than ONLINE/OFFLINE/UNKNOWN should
        cause an exception.
        """
        self.patch(FakeNetworkManagerState, "connection_state",
                   NM_STATE_CONNECTED_GLOBAL)
        yield self.assertFailure(is_machine_connected(), NetworkFailException)


class NetworkManagerBaseTestCase(MockerTestCase):
    """Base test case for NM state tests."""

    def setUp(self):
        """Setup the dbus object tree."""
        super(NetworkManagerBaseTestCase, self).setUp()

        self.dbusmock = self.mocker.mock()
        self.dbusmock.SystemBus()
        sysbusmock = self.mocker.mock()
        self.mocker.result(sysbusmock)

        sysbusmock.get_object(ARGS, KWARGS)

    def connect_proxy(self, exc=None):
        """Get a proxy mock object, and allow failing with specified exc."""
        proxymock = self.mocker.mock()
        self.mocker.result(proxymock)

        self.dbusmock.Interface(proxymock, ANY)
        ifmock = self.mocker.mock()
        self.mocker.result(ifmock)

        ifmock.connect_to_signal(ARGS, KWARGS)
        signalmock = self.mocker.mock()
        self.mocker.result(signalmock)

        proxymock.Get(ARGS, KWARGS)
        if exc is not None:
            self.mocker.result(exc)
        self.mocker.replay()

    # Invalid name "assert[A-Z].*"
    # pylint: disable=C0103
    def assertOnline(self, state):
        """Check that the state given is ONLINE."""
        self.assertEquals(state, ONLINE)

    def assertOffline(self, state):
        """Check that the state given is OFFLINE."""
        self.assertEquals(state, OFFLINE)

    def assertUnknown(self, state):
        """Check that the state was UNKNOWN."""
        self.assertEquals(state, UNKNOWN)
    # pylint: enable=C0103

    def get_nms(self, callback):
        """Get the NetworkManagerState object."""
        nms = NetworkManagerState(callback, self.dbusmock)
        nms.find_online_state()
        return nms

    def check_nm_error(self, callback, error):
        """Check that the error handling is correct."""
        nms = self.get_nms(callback)
        nms.got_error(error)

    def check_nm_state(self, callback, state):
        """Check the state handling is correct."""
        nms = self.get_nms(callback)
        nms.got_state(state)

    def check_nm_state_change(self, callback, fmstate, tostate):
        """Check the state change handling is correct."""
        nms = self.get_nms(callback)
        nms.got_state(fmstate)
        nms.state_changed(tostate)


class NetworkManagerStateTestCase(NetworkManagerBaseTestCase):
    """Test NetworkManager state retrieval code."""

    def setUp(self):
        """Setup the dbus object tree."""
        super(NetworkManagerStateTestCase, self).setUp()
        self.connect_proxy()

    def test_nm_asleep(self):
        """Asleep status should mean offline."""
        self.check_nm_state(self.assertOffline, NM_STATE_ASLEEP)

    def test_nm_asleep_old(self):
        """Asleep, old status, should mean offline."""
        self.check_nm_state(self.assertOffline, NM_STATE_ASLEEP_OLD)

    def test_nm_unknown(self):
        """Unknown status should be treated the same as OFFLINE."""
        self.check_nm_state(self.assertOffline, NM_STATE_UNKNOWN)

    def test_nm_online_old(self):
        """Check the connected, old status, case."""
        self.check_nm_state(self.assertOnline, NM_STATE_CONNECTED_OLD)

    def test_nm_offline_local(self):
        """Check the connected, local status, case."""
        self.check_nm_state(self.assertOffline, NM_STATE_CONNECTED_LOCAL)

    def test_nm_offline_site(self):
        """Check the connected, site status, case."""
        self.check_nm_state(self.assertOffline, NM_STATE_CONNECTED_SITE)

    def test_nm_online_global(self):
        """Check the connected, global status, case."""
        self.check_nm_state(self.assertOnline, NM_STATE_CONNECTED_GLOBAL)

    def test_nm_offline_old(self):
        """Check the disconnected, old status, case."""
        self.check_nm_state(self.assertOffline, NM_STATE_DISCONNECTED_OLD)

    def test_nm_offline(self):
        """Check the disconnected case."""
        self.check_nm_state(self.assertOffline, NM_STATE_DISCONNECTED)

    def test_nm_connecting_then_online_old(self):
        """Check the waiting for connection, old status, case."""
        self.check_nm_state_change(self.assertOnline,
                                   NM_STATE_CONNECTING_OLD,
                                   NM_STATE_CONNECTED_OLD)

    def test_nm_connecting_then_online(self):
        """Check the waiting for connection case."""
        self.check_nm_state_change(self.assertOnline,
                                   NM_STATE_CONNECTING,
                                   NM_STATE_CONNECTED_GLOBAL)

    def test_nm_connecting_then_offline_old(self):
        """Check the waiting but fail, old status, case."""
        self.check_nm_state_change(self.assertOffline,
                                   NM_STATE_CONNECTING_OLD,
                                   NM_STATE_DISCONNECTED_OLD)

    def test_nm_connecting_then_offline(self):
        """Check the waiting but fail case."""
        self.check_nm_state_change(self.assertOffline,
                                   NM_STATE_CONNECTING, NM_STATE_DISCONNECTED)


class NetworkManagerStateErrorsTestCase(NetworkManagerBaseTestCase):
    """Test NetworkManager state retrieval code."""

    def mock_except_while_getting_proxy(self, exc):
        """Simulate an exception while getting the DBus proxy object."""
        self.mocker.throw(exc)
        self.mocker.result(exc)
        self.mocker.replay()

    def test_nm_check_errors(self):
        """Trying to reach NM fails with some error."""
        self.connect_proxy(Exception)
        self.check_nm_error(self.assertOnline, Exception())

    def test_dbus_problem(self):
        """Check the case when DBus throws some other exception."""
        self.mock_except_while_getting_proxy(TestException)
        self.get_nms(self.assertOnline)
